"""
Module providing the base `CodeObject` and related functions.
"""

import collections
import copy
import platform

from brian2.core.base import weakproxy_with_fallback
from brian2.core.functions import DEFAULT_FUNCTIONS, Function
from brian2.core.names import Nameable
from brian2.equations.unitcheck import check_units_statements
from brian2.utils.logger import get_logger
from brian2.utils.stringtools import code_representation, indent

from .translation import analyse_identifiers

__all__ = ["CodeObject", "constant_or_scalar"]

logger = get_logger(__name__)


#: Dictionary with basic information about the current system (OS, etc.)
sys_info = {
    "system": platform.system(),
    "architecture": platform.architecture(),
    "machine": platform.machine(),
}


def constant_or_scalar(varname, variable):
    """
    Convenience function to generate code to access the value of a variable.
    Will return ``'varname'`` if the ``variable`` is a constant, and
    ``array_name[0]`` if it is a scalar array.
    """
    from brian2.devices.device import get_device  # avoid circular import

    if variable.array:
        return f"{get_device().get_array_name(variable)}[0]"
    else:
        return f"{varname}"


class CodeObject(Nameable):
    """
    Executable code object.

    The ``code`` can either be a string or a
    `brian2.codegen.templates.MultiTemplate`.

    After initialisation, the code is compiled with the given namespace
    using ``code.compile(namespace)``.

    Calling ``code(key1=val1, key2=val2)`` executes the code with the given
    variables inserted into the namespace.
    """

    #: The `CodeGenerator` class used by this `CodeObject`
    generator_class = None
    #: A short name for this type of `CodeObject`
    class_name = None

    def __init__(
        self,
        owner,
        code,
        variables,
        variable_indices,
        template_name,
        template_source,
        compiler_kwds,
        name="codeobject*",
    ):
        Nameable.__init__(self, name=name)
        self.owner = weakproxy_with_fallback(owner)
        self.code = code
        self.compiled_code = {}
        self.variables = variables
        self.variable_indices = variable_indices
        self.template_name = template_name
        self.template_source = template_source
        self.compiler_kwds = compiler_kwds

    def __getstate__(self):
        state = self.__dict__.copy()
        state["owner"] = self.owner.__repr__.__self__
        # Replace Function objects for standard functions by their name
        state["variables"] = self.variables.copy()
        for k, v in state["variables"].items():
            if isinstance(v, Function) and v is DEFAULT_FUNCTIONS[k]:
                state["variables"][k] = k
        return state

    def __setstate__(self, state):
        state["owner"] = weakproxy_with_fallback(state["owner"])
        for k, v in state["variables"].items():
            if isinstance(v, str):
                state["variables"][k] = DEFAULT_FUNCTIONS[k]
        self.__dict__ = state

    @classmethod
    def is_available(cls):
        """
        Whether this target for code generation is available. Should use a
        minimal example to check whether code generation works in general.
        """
        raise NotImplementedError(
            f"CodeObject class {cls.__name__} is missing an 'is_available' method."
        )

    def update_namespace(self):
        """
        Update the namespace for this timestep. Should only deal with variables
        where *the reference* changes every timestep, i.e. where the current
        reference in `namespace` is not correct.
        """
        pass

    def compile_block(self, block):
        raise NotImplementedError("Implement compile_block method")

    def compile(self):
        for block in ["before_run", "run", "after_run"]:
            self.compiled_code[block] = self.compile_block(block)

    def __call__(self, **kwds):
        self.update_namespace()
        self.namespace.update(**kwds)

        return self.run()

    def run_block(self, block):
        raise NotImplementedError("Implement run_block method")

    def before_run(self):
        """
        Runs the preparation code in the namespace. This code will only be
        executed once per run.

        Returns
        -------
        return_value : dict
            A dictionary with the keys corresponding to the `output_variables`
            defined during the call of `CodeGenerator.code_object`.
        """
        return self.run_block("before_run")

    def run(self):
        """
        Runs the main code in the namespace.

        Returns
        -------
        return_value : dict
            A dictionary with the keys corresponding to the `output_variables`
            defined during the call of `CodeGenerator.code_object`.
        """
        return self.run_block("run")

    def after_run(self):
        """
        Runs the finalizing code in the namespace. This code will only be
        executed once per run.

        Returns
        -------
        return_value : dict
            A dictionary with the keys corresponding to the `output_variables`
            defined during the call of `CodeGenerator.code_object`.
        """
        return self.run_block("after_run")


def _error_msg(code, name):
    """
    Little helper function for error messages.
    """
    error_msg = f"Error generating code for code object '{name}' "
    code_lines = [line for line in code.split("\n") if len(line.strip())]
    # If the abstract code is only one line, display it in full
    if len(code_lines) <= 1:
        error_msg += f"from this abstract code: '{code_lines[0]}'\n"
    else:
        error_msg += (
            f"from {len(code_lines)} lines of abstract code, first line is: "
            "'code_lines[0]'\n"
        )
    return error_msg


def check_compiler_kwds(compiler_kwds, accepted_kwds, target):
    """
    Internal function to check the provided compiler keywords against the list
    of understood keywords.

    Parameters
    ----------
    compiler_kwds : dict
        Dictionary of compiler keywords and respective list of values.
    accepted_kwds : list of str
        The compiler keywords understood by the code generation target
    target : str
        The name of the code generation target (used for the error message).

    Raises
    ------
    ValueError
        If a compiler keyword is not understood
    """
    for key in compiler_kwds:
        if key not in accepted_kwds:
            formatted_kwds = ", ".join(f"'{kw}'" for kw in accepted_kwds)
            raise ValueError(
                f"The keyword argument '{key}' is not understood by "
                f"the code generation target '{target}'. The valid "
                f"arguments are: {formatted_kwds}."
            )


def _merge_compiler_kwds(list_of_kwds):
    """
    Merges a list of keyword dictionaries. Values in these dictionaries are
    lists of values, the merged dictionaries will contain the concatenations
    of lists specified for the same key.

    Parameters
    ----------
    list_of_kwds : list of dict
        A list of compiler keyword dictionaries that should be merged.

    Returns
    -------
    merged_kwds : dict
        The merged dictionary
    """
    merged_kwds = collections.defaultdict(list)
    for kwds in list_of_kwds:
        for key, values in kwds.items():
            if not isinstance(values, list):
                raise TypeError(
                    f"Compiler keyword argument '{key}' requires a list of values."
                )
            merged_kwds[key].extend(values)
    return merged_kwds


def _gather_compiler_kwds(function, codeobj_class):
    """
    Gather all the compiler keywords for a function and its dependencies.

    Parameters
    ----------
    function : `Function`
        The function for which the compiler keywords should be gathered
    codeobj_class : type
        The class of `CodeObject` to use

    Returns
    -------
    kwds : dict
        A dictionary with the compiler arguments, a list of values for each
        key.
    """
    implementation = function.implementations[codeobj_class]
    all_kwds = [implementation.compiler_kwds]
    if implementation.dependencies is not None:
        for dependency in implementation.dependencies.values():
            all_kwds.append(_gather_compiler_kwds(dependency, codeobj_class))
    return _merge_compiler_kwds(all_kwds)


def create_runner_codeobj(
    group,
    code,
    template_name,
    run_namespace,
    user_code=None,
    variable_indices=None,
    name=None,
    check_units=True,
    needed_variables=None,
    additional_variables=None,
    template_kwds=None,
    override_conditional_write=None,
    codeobj_class=None,
):
    """Create a `CodeObject` for the execution of code in the context of a
    `Group`.

    Parameters
    ----------
    group : `Group`
        The group where the code is to be run
    code : str or dict of str
        The code to be executed.
    template_name : str
        The name of the template to use for the code.
    run_namespace : dict-like
        An additional namespace that is used for variable lookup (either
        an explicitly defined namespace or one taken from the local
        context).
    user_code : str, optional
        The code that had been specified by the user before other code was
        added automatically. If not specified, will be assumed to be identical
        to ``code``.
    variable_indices : dict-like, optional
        A mapping from `Variable` objects to index names (strings).  If none is
        given, uses the corresponding attribute of `group`.
    name : str, optional
        A name for this code object, will use ``group + '_codeobject*'`` if
        none is given.
    check_units : bool, optional
        Whether to check units in the statement. Defaults to ``True``.
    needed_variables: list of str, optional
        A list of variables that are neither present in the abstract code, nor
        in the ``USES_VARIABLES`` statement in the template. This is only
        rarely necessary, an example being a `StateMonitor` where the
        names of the variables are neither known to the template nor included
        in the abstract code statements.
    additional_variables : dict-like, optional
        A mapping of names to `Variable` objects, used in addition to the
        variables saved in `group`.
    template_kwds : dict, optional
        A dictionary of additional information that is passed to the template.
    override_conditional_write: list of str, optional
        A list of variable names which are used as conditions (e.g. for
        refractoriness) which should be ignored.
    codeobj_class : class, optional
        The `CodeObject` class to run code with. If not specified, defaults to
        the `group`'s ``codeobj_class`` attribute.
    """

    if name is None:
        if group is not None:
            name = f"{group.name}_{template_name}_codeobject*"
        else:
            name = f"{template_name}_codeobject*"

    if user_code is None:
        user_code = code

    if isinstance(code, str):
        code = {None: code}
        user_code = {None: user_code}

    msg = (
        f"Creating code object (group={group.name}, template name={template_name}) for"
        " abstract code:\n"
    )
    msg += indent(code_representation(code))
    logger.diagnostic(msg)
    from brian2.devices import get_device

    device = get_device()

    if override_conditional_write is None:
        override_conditional_write = set()
    else:
        override_conditional_write = set(override_conditional_write)

    if codeobj_class is None:
        codeobj_class = device.code_object_class(group.codeobj_class)
    else:
        codeobj_class = device.code_object_class(codeobj_class)

    template = getattr(codeobj_class.templater, template_name)
    template_variables = getattr(template, "variables", None)

    all_variables = dict(group.variables)
    if additional_variables is not None:
        all_variables.update(additional_variables)

    # Determine the identifiers that were used
    identifiers = set()
    user_identifiers = set()
    for v, u_v in zip(code.values(), user_code.values()):
        _, uk, u = analyse_identifiers(v, all_variables, recursive=True)
        identifiers |= uk | u
        _, uk, u = analyse_identifiers(u_v, all_variables, recursive=True)
        user_identifiers |= uk | u

    # Add variables that are not in the abstract code, nor specified in the
    # template but nevertheless necessary
    if needed_variables is None:
        needed_variables = []
    # Resolve all variables (variables used in the code and variables needed by
    # the template)
    variables = group.resolve_all(
        sorted(identifiers | set(needed_variables) | set(template_variables)),
        # template variables are not known to the user:
        user_identifiers=user_identifiers,
        additional_variables=additional_variables,
        run_namespace=run_namespace,
    )
    # We raise this error only now, because there is some non-obvious code path
    # where Jinja tries to get a Synapse's "name" attribute via syn['name'],
    # which then triggers the use of the `group_get_indices` template which does
    # not exist for standalone. Putting the check for template == None here
    # means we will first raise an error about the unknown identifier which will
    # then make Jinja try syn.name
    if template is None:
        codeobj_class_name = codeobj_class.class_name or codeobj_class.__name__
        raise AttributeError(
            f"'{codeobj_class_name}' does not provide a code "
            f"generation template '{template_name}'"
        )

    conditional_write_variables = {}
    # Add all the "conditional write" variables
    for var in variables.values():
        cond_write_var = getattr(var, "conditional_write", None)
        if cond_write_var in override_conditional_write:
            continue
        if cond_write_var is not None:
            if (
                cond_write_var.name in variables
                and not variables[cond_write_var.name] is cond_write_var
            ):
                logger.diagnostic(
                    f"Variable '{cond_write_var.name}' is needed for the "
                    "conditional write mechanism of variable "
                    f"'{var.name}'. Its name is already used for "
                    f"{variables[cond_write_var.name]!r}."
                )
            else:
                conditional_write_variables[cond_write_var.name] = cond_write_var

    variables.update(conditional_write_variables)

    if check_units:
        for c in code.values():
            # This is the first time that the code is parsed, catch errors
            try:
                check_units_statements(c, variables)
            except (SyntaxError, ValueError) as ex:
                error_msg = _error_msg(c, name)
                raise ValueError(error_msg) from ex

    all_variable_indices = copy.copy(group.variables.indices)
    if additional_variables is not None:
        all_variable_indices.update(additional_variables.indices)
    if variable_indices is not None:
        all_variable_indices.update(variable_indices)

    # Make "conditional write" variables use the same index as the variable
    # that depends on them
    for varname, var in variables.items():
        cond_write_var = getattr(var, "conditional_write", None)
        if cond_write_var is not None:
            all_variable_indices[cond_write_var.name] = all_variable_indices[varname]

    # Check that all functions are available
    for varname, value in variables.items():
        if isinstance(value, Function):
            try:
                value.implementations[codeobj_class]
            except KeyError as ex:
                # if we are dealing with numpy, add the default implementation
                from brian2.codegen.runtime.numpy_rt import NumpyCodeObject

                if codeobj_class is NumpyCodeObject:
                    value.implementations.add_numpy_implementation(value.pyfunc)
                else:
                    raise NotImplementedError(
                        f"Cannot use function '{varname}': {ex}"
                    ) from ex

    # Gather the additional compiler arguments declared by function
    # implementations
    all_keywords = [
        _gather_compiler_kwds(var, codeobj_class)
        for var in variables.values()
        if isinstance(var, Function)
    ]
    compiler_kwds = _merge_compiler_kwds(all_keywords)

    # Add the indices needed by the variables
    for varname in list(variables):
        var_index = all_variable_indices[varname]
        if var_index not in ("_idx", "0"):
            variables[var_index] = all_variables[var_index]

    return device.code_object(
        owner=group,
        name=name,
        abstract_code=code,
        variables=variables,
        template_name=template_name,
        variable_indices=all_variable_indices,
        template_kwds=template_kwds,
        codeobj_class=codeobj_class,
        override_conditional_write=override_conditional_write,
        compiler_kwds=compiler_kwds,
    )



"""
This script is used to ask for the CPU flags on Windows. We use this instead of
importing the cpuinfo package, because recent versions of py-cpuinfo use the
multiprocessing module, and any import of cpuinfo that is not within a
`if __name__ == '__main__':` block will lead to the script being executed twice.

The CPU flags are printed to stdout encoded as JSON.
"""

import json

if __name__ == "__main__":
    import cpuinfo

    flags = cpuinfo.get_cpu_info()["flags"]
    print(json.dumps(flags))



import pytest
from numpy.testing import assert_array_equal, assert_equal

from brian2 import *
from brian2.utils.logger import catch_logs


@pytest.mark.codegen_independent
def test_clock_attributes():
    clock = Clock(dt=1 * ms)
    assert_array_equal(clock.t, 0 * second)
    assert_array_equal(clock.timestep, 0)
    assert_array_equal(clock.dt, 1 * ms)


@pytest.mark.codegen_independent
def test_clock_dt_change():
    clock = Clock(dt=1 * ms)
    # at time 0s, all dt changes should be allowed
    clock.dt = 0.75 * ms
    clock._set_t_update_dt()
    clock.dt = 2.5 * ms
    clock._set_t_update_dt()
    clock.dt = 1 * ms
    clock._set_t_update_dt()

    # at 0.1ms only changes that are still representable as an integer of the
    # current time 1s are allowed
    clock.dt = 0.1 * ms
    clock._set_t_update_dt()

    clock.dt = 0.05 * ms
    clock._set_t_update_dt(target_t=0.1 * ms)
    clock.dt = 0.1 * ms
    clock._set_t_update_dt(target_t=0.1 * ms)
    clock.dt = 0.3 * ms
    with pytest.raises(ValueError):
        clock._set_t_update_dt(target_t=0.1 * ms)


@pytest.mark.codegen_independent
def test_defaultclock():
    defaultclock.dt = 1 * ms
    assert_equal(defaultclock.dt, 1 * ms)
    assert defaultclock.name == "defaultclock"


@pytest.mark.codegen_independent
def test_set_interval_warning():
    clock = Clock(dt=0.1 * ms)
    with catch_logs() as logs:
        clock.set_interval(0 * second, 1000 * second)  # no problem
    assert len(logs) == 0
    with catch_logs() as logs:
        clock.set_interval(0 * second, 10000000000 * second)  # too long
    assert len(logs) == 1
    assert logs[0][1].endswith("many_timesteps")


if __name__ == "__main__":
    test_clock_attributes()
    restore_initial_state()
    test_clock_dt_change()
    restore_initial_state()
    test_defaultclock()
    test_set_interval_warning()



"""
Spike-based adaptation of Licklider's model of pitch processing (autocorrelation with
delay lines) with phase locking.
"""
from brian2 import *

defaultclock.dt = .02 * ms

# Ear and sound
max_delay = 20*ms # 50 Hz
tau_ear = 1*ms
sigma_ear = 0.0
eqs_ear = '''
dx/dt = (sound-x)/tau_ear+0.1*(2./tau_ear)**.5*xi : 1 (unless refractory)
sound = 5*sin(2*pi*frequency*t)**3 : 1 # nonlinear distortion
#sound = 5*(sin(4*pi*frequency*t)+.5*sin(6*pi*frequency*t)) : 1 # missing fundamental
frequency = (200+200*t*Hz)*Hz : Hz # increasing pitch
'''
receptors = NeuronGroup(2, eqs_ear, threshold='x>1', reset='x=0',
                        refractory=2*ms, method='euler')
# Coincidence detectors
min_freq = 50*Hz
max_freq = 1000*Hz
num_neurons = 300
tau = 1*ms
sigma = .1
eqs_neurons = '''
dv/dt = -v/tau+sigma*(2./tau)**.5*xi : 1
'''

neurons = NeuronGroup(num_neurons, eqs_neurons, threshold='v>1', reset='v=0',
                      method='euler')

synapses = Synapses(receptors, neurons, on_pre='v += 0.5')
synapses.connect()
synapses.delay = 'i*1.0/exp(log(min_freq/Hz)+(j*1.0/(num_neurons-1))*log(max_freq/min_freq))*second'

spikes = SpikeMonitor(neurons)

run(500*ms)
plot(spikes.t/ms, spikes.i, '.k')
xlabel('Time (ms)')
ylabel('Frequency')
yticks([0, 99, 199, 299],
       array(1. / synapses.delay[1, [0, 99, 199, 299]], dtype=int))
show()



# coding=utf-8
"""
Modeling neuron-glia interactions with the Brian 2 simulator
Marcel Stimberg, Dan F. M. Goodman, Romain Brette, Maurizio De Pittà
bioRxiv 198366; doi: https://doi.org/10.1101/198366

Figure 3: Modeling of modulation of synaptic release by gliotransmission.

Three synapses: the first one without astrocyte, the remaining two respectively with
open-loop and close-loop gliotransmission (see De Pitta' et al., 2011, 2016)
"""
from brian2 import *

import plot_utils as pu

set_device('cpp_standalone', directory=None)  # Use fast "C++ standalone mode"

################################################################################
# Model parameters
################################################################################
### General parameters
transient = 16.5*second
duration = transient + 600*ms   # Total simulation time
sim_dt = 1*ms                   # Integrator/sampling step

### Synapse parameters
rho_c = 0.005                   # Synaptic vesicle-to-extracellular space volume ratio
Y_T = 500*mmolar                # Total vesicular neurotransmitter concentration
Omega_c = 40/second             # Neurotransmitter clearance rate
U_0__star = 0.6                 # Resting synaptic release probability
Omega_f = 3.33/second           # Synaptic facilitation rate
Omega_d = 2.0/second            # Synaptic depression rate
# --- Presynaptic receptors
O_G = 1.5/umolar/second         # Agonist binding (activating) rate
Omega_G = 0.5/(60*second)       # Agonist release (deactivating) rate

### Astrocyte parameters
# ---  Calcium fluxes
O_P = 0.9*umolar/second         # Maximal Ca^2+ uptake rate by SERCAs
K_P = 0.05 * umolar             # Ca2+ affinity of SERCAs
C_T = 2*umolar                  # Total cell free Ca^2+ content
rho_A = 0.18                    # ER-to-cytoplasm volume ratio
Omega_C = 6/second              # Maximal rate of Ca^2+ release by IP_3Rs
Omega_L = 0.1/second            # Maximal rate of Ca^2+ leak from the ER
# --- IP_3R kinectics
d_1 = 0.13*umolar               # IP_3 binding affinity
d_2 = 1.05*umolar               # Ca^2+ inactivation dissociation constant
O_2 = 0.2/umolar/second         # IP_3R binding rate for Ca^2+ inhibition
d_3 = 0.9434*umolar             # IP_3 dissociation constant
d_5 = 0.08*umolar               # Ca^2+ activation dissociation constant
# --- IP_3 production
O_delta = 0.6*umolar/second     # Maximal rate of IP_3 production by PLCdelta
kappa_delta = 1.5* umolar       # Inhibition constant of PLC_delta by IP_3
K_delta = 0.1*umolar            # Ca^2+ affinity of PLCdelta
# --- IP_3 degradation
Omega_5P = 0.05/second          # Maximal rate of IP_3 degradation by IP-5P
K_D = 0.7*umolar                # Ca^2+ affinity of IP3-3K
K_3K = 1.0*umolar               # IP_3 affinity of IP_3-3K
O_3K = 4.5*umolar/second        # Maximal rate of IP_3 degradation by IP_3-3K
# --- IP_3 diffusion
F_ex = 2.0*umolar/second        # Maximal exogenous IP3 flow
I_Theta = 0.3*umolar            # Threshold gradient for IP_3 diffusion
omega_I = 0.05*umolar           # Scaling factor of diffusion
# --- Gliotransmitter release and time course
C_Theta = 0.5*umolar            # Ca^2+ threshold for exocytosis
Omega_A = 0.6/second            # Gliotransmitter recycling rate
U_A = 0.6                       # Gliotransmitter release probability
G_T = 200*mmolar                # Total vesicular gliotransmitter concentration
rho_e = 6.5e-4                  # Astrocytic vesicle-to-extracellular volume ratio
Omega_e = 60/second             # Gliotransmitter clearance rate
alpha = 0.0                     # Gliotransmission nature

################################################################################
# Model definition
################################################################################
defaultclock.dt = sim_dt  # Set the integration time

### "Neurons"
# We are only interested in the activity of the synapse, so we replace the
# neurons by trivial "dummy" groups
spikes = [0, 50, 100, 150, 200,
          300, 310, 320, 330, 340, 350, 360, 370, 380, 390, 400]*ms
spikes += transient  # allow for some initial transient
source_neurons = SpikeGeneratorGroup(1, np.zeros(len(spikes)), spikes)
target_neurons = NeuronGroup(1, '')

### Synapses
# Note that the synapse does not actually have any effect on the post-synaptic
# target
# Also note that for easier plotting we do not use the "event-driven" flag here,
# even though the value of u_S and x_S only needs to be updated on the arrival
# of a spike
synapses_eqs = '''
# Neurotransmitter
dY_S/dt = -Omega_c * Y_S        : mmolar (clock-driven)
# Fraction of activated presynaptic receptors
dGamma_S/dt = O_G * G_A * (1 - Gamma_S) - 
              Omega_G * Gamma_S : 1 (clock-driven)
# Usage of releasable neurotransmitter per single action potential:
du_S/dt = -Omega_f * u_S        : 1 (clock-driven)
# Fraction of synaptic neurotransmitter resources available:
dx_S/dt = Omega_d *(1 - x_S)    : 1 (clock-driven)
# released synaptic neurotransmitter resources:
r_S                             : 1
# gliotransmitter concentration in the extracellular space:
G_A                             : mmolar
'''
synapses_action = '''
U_0 = (1 - Gamma_S) * U_0__star + alpha * Gamma_S
u_S += U_0 * (1 - u_S)
r_S = u_S * x_S
x_S -= r_S
Y_S += rho_c * Y_T * r_S
'''
synapses = Synapses(source_neurons, target_neurons,
                    model=synapses_eqs, on_pre=synapses_action,
                    method='exact')
# We create three synapses, only the second and third ones are modulated by astrocytes
synapses.connect(True, n=3)

### Astrocytes
# The astrocyte emits gliotransmitter when its Ca^2+ concentration crosses
# a threshold
astro_eqs = '''
# IP_3 dynamics:
dI/dt = J_delta - J_3K - J_5P + J_ex                             : mmolar
J_delta = O_delta/(1 + I/kappa_delta) * C**2/(C**2 + K_delta**2) : mmolar/second
J_3K = O_3K * C**4/(C**4 + K_D**4) * I/(I + K_3K)                : mmolar/second
J_5P = Omega_5P*I                                                : mmolar/second
# Exogenous stimulation
delta_I_bias = I - I_bias          : mmolar
J_ex = -F_ex/2*(1 + tanh((abs(delta_I_bias) - I_Theta)/omega_I)) *
                sign(delta_I_bias) : mmolar/second
I_bias                             : mmolar (constant)

# Ca^2+-induced Ca^2+ release:
dC/dt = (Omega_C * m_inf**3 * h**3 + Omega_L) * (C_T - (1 + rho_A)*C) -
        O_P * C**2/(C**2 + K_P**2) : mmolar
dh/dt = (h_inf - h)/tau_h : 1  # IP3R de-inactivation probability
m_inf = I/(I + d_1) * C/(C + d_5)  : 1
h_inf = Q_2/(Q_2 + C)              : 1
tau_h = 1/(O_2 * (Q_2 + C))        : second
Q_2 = d_2 * (I + d_1)/(I + d_3)    : mmolar
# Fraction of gliotransmitter resources available:
dx_A/dt = Omega_A * (1 - x_A)      : 1
# gliotransmitter concentration in the extracellular space:
dG_A/dt = -Omega_e*G_A             : mmolar
'''
glio_release = '''
G_A += rho_e * G_T * U_A * x_A
x_A -= U_A *  x_A
'''
# The following formulation makes sure that a "spike" is only triggered at the
# first threshold crossing -- the astrocyte is considered "refractory" (i.e.,
# not allowed to trigger another event) as long as the Ca2+ concentration
# remains above threshold
# The gliotransmitter release happens when the threshold is crossed, in Brian
# terms it can therefore be considered a "reset"
astrocyte = NeuronGroup(2, astro_eqs,
                        threshold='C>C_Theta',
                        refractory='C>C_Theta',
                        reset=glio_release,
                        method='rk4')
# Different length of stimulation
astrocyte.x_A = 1.0
astrocyte.h = 0.9
astrocyte.I = 0.4*umolar
astrocyte.I_bias = np.asarray([0.8, 1.25])*umolar

# Connection between astrocytes and the second synapse. Note that in this
# special case, where the synapse is only influenced by the gliotransmitter from
# a single astrocyte, the '(linked)' variable mechanism could be used instead.
# The mechanism used below is more general and can add the contribution of
# several astrocytes
ecs_astro_to_syn = Synapses(astrocyte, synapses,
                            'G_A_post = G_A_pre : mmolar (summed)')
# Connect second and third synapse to a different astrocyte
ecs_astro_to_syn.connect(j='i+1')

################################################################################
# Monitors
################################################################################
# Note that we cannot use "record=True" for synapses in C++ standalone mode --
# the StateMonitor needs to know the number of elements to record from during
# its initialization, but in C++ standalone mode, no synapses have been created
# yet. We therefore explicitly state to record from the three synapses.
syn_mon = StateMonitor(synapses, variables=['u_S', 'x_S', 'r_S', 'Y_S'],
                       record=[0, 1, 2])
ast_mon = StateMonitor(astrocyte, variables=['C', 'G_A'], record=True)

################################################################################
# Simulation run
################################################################################
run(duration, report='text')

################################################################################
# Analysis and plotting
################################################################################
from matplotlib import cycler
plt.style.use('figures.mplstyle')

fig, ax = plt.subplots(nrows=7, ncols=1, figsize=(6.26894, 6.26894 * 1.2),
                       gridspec_kw={'height_ratios': [3, 2, 1, 1, 3, 3, 3],
                                    'top': 0.98, 'bottom': 0.08,
                                    'left': 0.15, 'right': 0.95})

## Ca^2+ traces of the two astrocytes
ax[0].plot((ast_mon.t-transient)/second, ast_mon.C[0]/umolar, '-', color='C2')
ax[0].plot((ast_mon.t-transient)/second, ast_mon.C[1]/umolar, '-', color='C3')
## Add threshold for gliotransmitter release
ax[0].plot(np.asarray([-transient/second, 0.0]),
           np.asarray([C_Theta, C_Theta])/umolar, ':', color='gray')
ax[0].set(xlim=[-transient/second, 0.0], yticks=[0., 0.4, 0.8, 1.2],
          ylabel=r'$C$ ($\mu$M)')
pu.adjust_spines(ax[0], ['left'])

## Gliotransmitter concentration in the extracellular space
ax[1].plot((ast_mon.t-transient)/second, ast_mon.G_A[0]/umolar, '-', color='C2')
ax[1].plot((ast_mon.t-transient)/second, ast_mon.G_A[1]/umolar, '-', color='C3')
ax[1].set(yticks=[0., 50., 100.], xlim=[-transient/second, 0.0],
          xlabel='time (s)', ylabel=r'$G_A$ ($\mu$M)')
pu.adjust_spines(ax[1], ['left', 'bottom'])

## Turn off one axis to display x-labeling of ax[1] correctly
ax[2].axis('off')

## Synaptic stimulation
ax[3].vlines((spikes-transient)/ms, 0, 1, clip_on=False)
ax[3].set(xlim=(0, (duration-transient)/ms))
ax[3].axis('off')

## Synaptic variables
# Use a custom cycle that uses black as the first color
prop_cycle = cycler(color='k').concat(matplotlib.rcParams['axes.prop_cycle'][2:])
ax[4].set(xlim=(0, (duration-transient)/ms), ylim=[0., 1.],
          yticks=np.arange(0, 1.1, .25), ylabel='$u_S$',
          prop_cycle=prop_cycle)
ax[4].plot((syn_mon.t-transient)/ms, syn_mon.u_S.T)
pu.adjust_spines(ax[4], ['left'])

ax[5].set(xlim=(0, (duration-transient)/ms), ylim=[-0.05, 1.],
          yticks=np.arange(0, 1.1, .25), ylabel='$x_S$',
          prop_cycle=prop_cycle)
ax[5].plot((syn_mon.t-transient)/ms, syn_mon.x_S.T)
pu.adjust_spines(ax[5], ['left'])

ax[6].set(xlim=(0, (duration-transient)/ms), ylim=(-5., 1500),
          xticks=np.arange(0, (duration-transient)/ms, 100), xlabel='time (ms)',
          yticks=[0, 500, 1000, 1500], ylabel=r'$Y_S$ ($\mu$M)',
          prop_cycle=prop_cycle)
ax[6].plot((syn_mon.t-transient)/ms, syn_mon.Y_S.T/umolar)
ax[6].legend(['no gliotransmission',
              'weak gliotransmission',
              'stronger gliotransmission'], loc='upper right')
pu.adjust_spines(ax[6], ['left', 'bottom'])

pu.adjust_ylabels(ax, x_offset=-0.11)

plt.show()


'''
How long does synapse creation take? Brian1 for comparison
'''
import time
import pickle

import numpy as np
import joblib

from brian import *

repetitions = 5

memory = joblib.Memory(cachedir='.', verbose=0)

@memory.cache
def test_connectivity(N, i, j, n, p):
    G = NeuronGroup(N, '')
    # Do it once without measuring the time to ignore the compilation time for
    # C code -- not really necessary for Brian1 but leave it in for perfect
    # comparability
    S = Synapses(G, G, '')
    S[:, :] = i
    connections = len(S)
    del S
    times = []
    for _ in range(repetitions):
        S = Synapses(G, G, '')
        start = time.time()
        S[:, :] = i
        times.append(time.time() - start)
        del S

    return np.median(times), connections

conditions = [('Full', True),
              ('Full (no-self)', 'i != j'),
              ('One-to-one', 'i == j'),
              ('Simple neighbourhood', 'abs(i-j) < 5'),
              ('Gauss neighbourhood', 'exp(-(i - j)**2/5) > 0.005'),
              ('Random (50%)', 0.5),
              ('Random (10%)', 0.1),
              ('Random (1%)', 0.01),
              ('Random no-self (50%)', '(i != j) * 0.5'),
              ('Random no-self (10%)', '(i != j) * 0.1'),
              ('Random no-self (1%)', '(i != j) * 0.01')]
results = {}
lang_name = 'Brian 1'
max_connections = 10000000
for pattern, condition in conditions:
    N = 1
    connections = took = 0
    while connections < max_connections and took < 60.:
        print(lang_name, pattern)
        took, connections = test_connectivity(N, condition, None, 1, 1.)
        print(N, '%.4fs (for %d connections)' % (took, connections))
        results[(lang_name, connections, pattern)] = took
        N *= 2

with open('synapse_creation_times_brian1.pickle', 'w') as f:
    pickle.dump(results, f)



"""
    Automatically generate Brian's reference documentation.

    Based on sphinx-apidoc, published under a BSD license: http://sphinx-doc.org/
"""

import inspect
import os
import sys
from os import path

from .examplefinder import auto_find_examples

INITPY = "__init__.py"

OPTIONS = ["show-inheritance"]


def makename(package, module):
    """Join package and module with a dot."""
    # Both package and module can be None/empty.
    if package:
        name = package
        if module:
            name += f".{module}"
    else:
        name = module
    return name


def write_file(name, text, destdir, suffix):
    """Write the output file for module/package <name>."""
    fname = path.join(destdir, f"{name}.{suffix}")
    print(f"Creating file {fname}.")
    f = open(fname, "w")
    try:
        f.write(text)
    finally:
        f.close()


def format_heading(level, text):
    """Create a heading of <level> [1, 2 or 3 supported]."""
    underlining = ["=", "-", "~"][level - 1] * len(text)
    return f"{text}\n{underlining}\n\n"


def format_directive(module, destdir, package=None, basename="brian2"):
    """Create the automodule directive and add the options."""
    directive = f".. automodule:: {makename(package, module)}\n"
    for option in OPTIONS:
        directive += f"    :{option}:\n"
    directive += "\n"
    # document all the classes in the modules
    full_name = f"{basename}.{module}"
    __import__(full_name)
    mod = sys.modules[full_name]
    dir_members = dir(mod)
    classes = []
    functions = []
    variables = []
    for member in dir_members:
        _temp = __import__(full_name, {}, {}, [member], 0)
        member_obj = getattr(_temp, member)
        member_module = getattr(member_obj, "__module__", None)
        # only document members that where defined in this module
        if member_module == full_name and not member.startswith("_"):
            if inspect.isclass(member_obj):
                classes.append((member, member_obj))
            elif inspect.isfunction(member_obj):
                functions.append((member, member_obj))
            else:
                variables.append((member, member_obj))

    if classes:
        directive += "**Classes**\n\n"
        for member, member_obj in classes:
            directive += f".. autosummary:: {member}\n"
            directive += "    :toctree:\n\n"
            create_member_file(full_name, member, member_obj, destdir)
    if functions:
        directive += "**Functions**\n\n"
        for member, member_obj in functions:
            directive += f".. autosummary:: {member}\n"
            directive += "    :toctree:\n\n"
            create_member_file(full_name, member, member_obj, destdir)
    if variables:
        directive += "**Objects**\n\n"
        for member, member_obj in variables:
            directive += f".. autosummary:: {member}\n"
            directive += "    :toctree:\n\n"
            create_member_file(full_name, member, member_obj, destdir)

    return directive


def find_shortest_import(module_name, obj_name):
    parts = module_name.split(".")
    for idx in range(1, len(parts) + 1):
        try:
            result = __import__(
                ".".join(parts[:idx]), globals(), {}, fromlist=[str(obj_name)], level=0
            )
            result_obj = getattr(result, obj_name, None)
            if (
                result_obj is not None
                and getattr(result_obj, "__module__", None) == module_name
            ):
                # import seems to have worked
                return ".".join(parts[:idx])
        except ImportError:
            pass
    raise AssertionError(f"Couldn't import {module_name}.{obj_name}")


def create_member_file(module_name, member, member_obj, destdir, suffix="rst"):
    """Build the text of the file and write the file."""

    text = f".. currentmodule:: {module_name}\n\n"

    shortest_import = find_shortest_import(module_name, member)
    import_text = f"(*Shortest import*: ``from {shortest_import} import {member})``\n\n"
    if inspect.isclass(member_obj):
        text += format_heading(1, f"{member} class")
        text += import_text
        text += f".. autoclass:: {member}\n\n"
        text += auto_find_examples(member_obj, headersymbol="-")
    elif inspect.isfunction(member_obj):
        text += format_heading(1, f"{member} function")
        text += import_text
        text += f".. autofunction:: {member}\n\n"
    else:
        text += format_heading(1, f"{member} object")
        text += import_text
        text += f".. autodata:: {member}\n"

    write_file(makename(module_name, member), text, destdir, suffix)


def create_package_file(
    root, master_package, subroot, py_files, subs, destdir, excludes, suffix="rst"
):
    """Build the text of the file and write the file."""
    package = path.split(root)[-1]
    text = format_heading(1, f"{package} package")
    # add each module in the package
    for py_file in py_files:
        if shall_skip(path.join(root, py_file)):
            continue
        is_package = py_file == INITPY
        py_file = path.splitext(py_file)[0]
        py_path = makename(subroot, py_file)
        # we don't want an additional header for the package,
        if not is_package:
            heading = f":mod:`{py_file}` module"
            text += format_heading(2, heading)
        text += format_directive(
            is_package and subroot or py_path, destdir, master_package
        )
        text += "\n"

    # build a list of directories that are packages (contain an INITPY file)
    subs = [sub for sub in subs if path.isfile(path.join(root, sub, INITPY))]
    # if there are some package directories, add a TOC for theses subpackages
    if subs:
        text += format_heading(2, "Subpackages")
        text += ".. toctree::\n"
        text += "    :maxdepth: 2\n\n"
        for sub in subs:
            if not is_excluded(os.path.join(root, sub), excludes):
                text += f"    {makename(master_package, subroot)}.{sub}\n"
        text += "\n"

    write_file(makename(master_package, subroot), text, destdir, suffix)


def shall_skip(module):
    """Check if we want to skip this module."""
    # skip it if there is nothing (or just \n or \r\n) in the file
    return path.getsize(module) <= 2


def recurse_tree(rootpath, exclude_dirs, exclude_files, destdir):
    """
    Look for every file in the directory tree and create the corresponding
    ReST files.
    """
    # use absolute path for root, as relative paths like '../../foo' cause
    # 'if "/." in root ...' to filter out *all* modules otherwise
    rootpath = path.normpath(path.abspath(rootpath))
    # check if the base directory is a package and get its name
    if INITPY in os.listdir(rootpath):
        root_package = rootpath.split(path.sep)[-1]
    else:
        # otherwise, the base is a directory with packages
        root_package = None

    toplevels = []
    for root, subs, files in os.walk(rootpath):
        if is_excluded(root, exclude_dirs):
            del subs[:]
            continue
        # document only Python module files
        py_files = sorted(
            [
                f
                for f in files
                if (path.splitext(f)[1] == ".py" and f not in exclude_files)
            ]
        )
        is_pkg = INITPY in py_files
        if is_pkg:
            py_files.remove(INITPY)
            py_files.insert(0, INITPY)
        elif root != rootpath:
            # only accept non-package at toplevel
            del subs[:]
            continue
        # remove hidden ('.') and private ('_') directories
        subs[:] = sorted(sub for sub in subs if sub[0] not in [".", "_"])

        if is_pkg:
            # we are in a package with something to document
            if subs or len(py_files) > 1 or not shall_skip(path.join(root, INITPY)):
                subpackage = (
                    root[len(rootpath) :].lstrip(path.sep).replace(path.sep, ".")
                )
                create_package_file(
                    root,
                    root_package,
                    subpackage,
                    py_files,
                    subs,
                    destdir,
                    exclude_dirs,
                )
                toplevels.append(makename(root_package, subpackage))
        else:
            raise AssertionError("Expected it to be a package")

    return toplevels


def normalize_excludes(rootpath, excludes):
    """
    Normalize the excluded directory list:
    * must be either an absolute path or start with rootpath,
    * otherwise it is joined with rootpath
    * with trailing slash
    """
    f_excludes = []
    for exclude in excludes:
        if not path.isabs(exclude) and not exclude.startswith(rootpath):
            exclude = path.join(rootpath, exclude)
        f_excludes.append(path.normpath(exclude) + path.sep)
    return f_excludes


def is_excluded(root, excludes):
    """
    Check if the directory is in the exclude list.

    Note: by having trailing slashes, we avoid common prefix issues, like
          e.g. an exlude "foo" also accidentally excluding "foobar".
    """
    sep = path.sep
    if not root.endswith(sep):
        root += sep
    for exclude in excludes:
        if root.startswith(exclude):
            return True
    return False


def main(rootpath, exclude_dirs, exclude_files, destdir):
    if not os.path.exists(destdir):
        os.makedirs(destdir)

    exclude_dirs = normalize_excludes(rootpath, exclude_dirs)
    recurse_tree(rootpath, exclude_dirs, exclude_files, destdir)



{# USES_VARIABLES { N } #}
{# ALLOWS_SCALAR_WRITE #}
{% extends 'common_group.pyx' %}

{% block maincode %}
    cdef size_t _N = {{constant_or_scalar('N', variables['N'])}}
    _vectorisation_idx = 1
    
    {{scalar_code['condition']|autoindent}}
    {{scalar_code['statement']|autoindent}}

    for _idx in range(_N):
        _vectorisation_idx = _idx
        
        {{vector_code['condition']|autoindent}}
        if _cond:
            # this looks silly but is necessary to avoid an error if there is
            # no vector code (setting a scalar variable)
            pass
            {{vector_code['statement']|autoindent}}

{% endblock %}



# -*- coding: utf-8 -*-
#
# Brian 2 documentation build configuration file, created by
# sphinx-quickstart on Tue Aug 14 14:57:16 2012.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys, os, shutil

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))


# Mock the scipy module
class Mock(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    @classmethod
    def __getattr__(cls, name):
        if name in ('__file__', '__path__'):
            return '/dev/null'
        elif name[0] == name[0].upper():
            mockType = type(name, (), {})
            mockType.__module__ = __name__
            return mockType
        else:
            return Mock()

MOCK_MODULES = ['brian', 'brian.hears',   # mock Brian1 for the bridge
                # Mock additional packages to reduce installation time on
                # readthedocs
                'Cython', 'Cython.Compiler', 'Cython.Build',
                'scipy']
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()

# Store the examples and tutorials directory in an environment variable,
# used by the "example finder" that is called as part of the autoclass documentation
os.environ['BRIAN2_DOCS_EXAMPLE_DIR'] = root_dir = os.path.abspath('../examples')
os.environ['BRIAN2_DOCS_TUTORIALS_DIR'] = root_dir = os.path.abspath('../tutorials')

if 'BRIAN2_DOCS_QUICK_REBUILD' not in os.environ:
    # -- Automatically generate the reference documentation ------------------------
    import brian2.sphinxext.generate_reference as generate_reference
    # first generate the reference documentation
    target_dir = './reference'
    if os.path.exists(target_dir):
        shutil.rmtree(target_dir)
    os.makedirs(target_dir)
    abs_root = os.path.abspath('../brian2')
    generate_reference.main(abs_root, exclude_dirs=['tests', 'sphinxext'],
                            exclude_files=['conftest.py'],
                            destdir=target_dir)
    
    # -- Automatically generate the examples documentation -------------------------
    import brian2.sphinxext.generate_examples as generate_examples
    target_dir = './examples'
    root_dir = os.path.abspath('../examples')
    generate_examples.main(root_dir, target_dir)

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.8'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.doctest',
              'sphinx.ext.intersphinx', 'sphinx.ext.todo',
              'sphinx.ext.coverage', 'sphinx.ext.mathjax',
              'sphinx.ext.viewcode',
              'brian2.sphinxext.briandoc',
              'sphinx.ext.autosummary',
              'sphinx.ext.extlinks',
              'sphinx_tabs.tabs']

# Set readthedocs theme if available
try:
    import sphinx_rtd_theme
    extensions += ['sphinx_rtd_theme']
    html_theme = 'sphinx_rtd_theme'
except ImportError:
    pass

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Brian 2'
import datetime, time
copyright = "2012–{}, Brian authors".format(
    datetime.datetime.fromtimestamp(
        int(os.environ.get("SOURCE_DATE_EPOCH", time.time())),
        tz=datetime.timezone.utc
    ).year
)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
from importlib.metadata import version

release = version('brian2')
# The short X.Y.Z version.
version = ".".join(release.split(".")[:3])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = 'brianobj'

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['brian2.']

sphinx_tabs_disable_tab_closing = True

# -- Options for HTML output ---------------------------------------------------

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}


# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/brian-logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Brian2doc'

# Suppress warnings about the mybinder badges
suppress_warnings = ['image.nonlocal_uri']

# -- Options for LaTeX output --------------------------------------------------
latex_engine = 'xelatex'

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', 'Brian2.tex', 'Brian 2 Documentation',
   'Brian authors', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'brian2', 'Brian 2 Documentation',
     ['Brian authors'], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'Brian2', 'Brian 2 Documentation',
   'Brian authors', 'Brian2', 'One line description of project.',
   'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'


intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy' : ('https://numpy.org/doc/stable', None),
    'scipy' : ('https://docs.scipy.org/doc/scipy', None),
    'sympy' : ('https://docs.sympy.org/dev/', None)
}

autodoc_default_options = {'show-inheritance': True}

doctest_global_setup = 'from brian2 import *'
highlight_language = 'python'  # instead of python3 (default for sphinx>=1.4)

# Configure linking to github
import sphinx
sphinx_version = tuple(int(x) for x in sphinx.__version__.split('.'))
if sphinx_version >= (4, 0, 0): 
    extlinks = {'issue': ('https://github.com/brian-team/brian2/issues/%s',
                          '# %s')}
else:
    extlinks = {'issue': ('https://github.com/brian-team/brian2/issues/%s',
                          '# ')}



"""
All Brian objects should derive from `BrianObject`.
"""

import functools
import os
import sys
import traceback
import weakref

from brian2.core.names import Nameable
from brian2.units.allunits import second
from brian2.units.fundamentalunits import check_units
from brian2.utils.logger import get_logger

__all__ = [
    "BrianObject",
    "BrianObjectException",
]

logger = get_logger(__name__)


class BrianObject(Nameable):
    """
    All Brian objects derive from this class, defines magic tracking and update.

    See the documentation for `Network` for an explanation of which
    objects get updated in which order.

    Parameters
    ----------
    dt : `Quantity`, optional
        The time step to be used for the simulation. Cannot be combined with
        the `clock` argument.
    clock : `Clock`, optional
        The update clock to be used. If neither a clock, nor the `dt` argument
        is specified, the `defaultclock` will be used.
    when : str, optional
        In which scheduling slot to simulate the object during a time step.
        Defaults to ``'start'``. See :ref:`scheduling` for possible values.
    order : int, optional
        The priority of this object for operations occurring at the same time
        step and in the same scheduling slot. Defaults to 0.
    namespace: dict, optional
        A dictionary mapping identifier names to objects. If not given, the
        namespace will be filled in at the time of the call of `Network.run`,
        with either the values from the ``namespace`` argument of the
        `Network.run` method or from the local context, if no such argument is
        given.
    name : str, optional
        A unique name for the object - one will be assigned automatically if
        not provided (of the form ``brianobject_1``, etc.).
    Notes
    -----

    The set of all `BrianObject` objects is stored in ``BrianObject.__instances__()``.
    """

    @check_units(dt=second)
    def __init__(
        self,
        dt=None,
        clock=None,
        when="start",
        order=0,
        namespace=None,
        name="brianobject*",
    ):
        # Setup traceback information for this object
        creation_stack = []
        bases = []
        for modulename in ["brian2"]:
            if modulename in sys.modules:
                base, _ = os.path.split(sys.modules[modulename].__file__)
                bases.append(base)
        for fname, linenum, funcname, line in traceback.extract_stack():
            if all(base not in fname for base in bases):
                s = f"  File '{fname}', line {linenum}, in {funcname}\n    {line}"
                creation_stack.append(s)
        creation_stack = [""] + creation_stack
        #: A string indicating where this object was created (traceback with any parts of Brian code removed)
        self._creation_stack = creation_stack[-1]
        self._full_creation_stack = "\n".join(creation_stack)

        if dt is not None and clock is not None:
            raise ValueError("Can only specify either a dt or a clock, not both.")

        if not isinstance(when, str):
            from brian2.core.clocks import Clock

            # Give some helpful error messages for users coming from the alpha
            # version
            if isinstance(when, Clock):
                raise TypeError(
                    "Do not use the 'when' argument for "
                    "specifying a clock, either provide a "
                    "timestep for the 'dt' argument or a Clock "
                    "object for 'clock'."
                )
            if isinstance(when, tuple):
                raise TypeError(
                    "Use the separate keyword arguments, 'dt' (or "
                    "'clock'), 'when', and 'order' instead of "
                    "providing a tuple for 'when'. Only use the "
                    "'when' argument for the scheduling slot."
                )
            # General error
            raise TypeError(
                "The 'when' argument has to be a string "
                "specifying the scheduling slot (e.g. 'start')."
            )

        Nameable.__init__(self, name)

        #: The clock used for simulating this object
        self._clock = clock
        if clock is None:
            from brian2.core.clocks import Clock, defaultclock

            if dt is not None:
                self._clock = Clock(dt=dt, name=self.name + "_clock*")
            else:
                self._clock = defaultclock

        if getattr(self._clock, "_is_proxy", False):
            from brian2.devices.device import get_device

            self._clock = get_device().defaultclock

        #: Used to remember the `Network` in which this object has been included
        #: before, to raise an error if it is included in a new `Network`
        self._network = None

        #: The ID string determining when the object should be updated in `Network.run`.
        self.when = when

        #: The order in which objects with the same clock and ``when`` should be updated
        self.order = order

        self._dependencies = set()
        self._contained_objects = []
        self._code_objects = []

        self._active = True

        #: The scope key is used to determine which objects are collected by magic
        self._scope_key = self._scope_current_key

        # Make sure that keys in the namespace are valid
        if namespace is None:
            # Do not overwrite namespace if already set (e.g. in StateMonitor)
            namespace = getattr(self, "namespace", {})
        for key in namespace:
            if key.startswith("_"):
                raise ValueError(
                    "Names starting with underscores are "
                    "reserved for internal use an cannot be "
                    "defined in the namespace argument."
                )
        #: The group-specific namespace
        self.namespace = namespace

        logger.diagnostic(
            f"Created BrianObject with name {self.name}, "
            f"clock={self._clock}, "
            f"when={self.when}, order={self.order}"
        )

    #: Global key value for ipython cell restrict magic
    _scope_current_key = 0

    #: Whether or not `MagicNetwork` is invalidated when a new `BrianObject` of this type is added
    invalidates_magic_network = True

    #: Whether or not the object should be added to a `MagicNetwork`. Note that
    #: all objects in `BrianObject.contained_objects` are automatically added
    #: when the parent object is added, therefore e.g. `NeuronGroup` should set
    #: `add_to_magic_network` to ``True``, but it should not be set for all the
    #: dependent objects such as `StateUpdater`
    add_to_magic_network = False

    def __del__(self):
        # For objects that get garbage collected, raise a warning if they have
        # never been part of a network
        if (
            getattr(self, "_network", "uninitialized") is None
            and getattr(self, "group", None) is None
        ):
            logger.warn(
                f"The object '{self.name}' is getting deleted, but was never included in a network. "
                "This probably means that you did not store the object reference in a variable, "
                "or that the variable was not used to construct the network.\n"
                "The object was created here (most recent call only):\n"
                + self._creation_stack,
                name_suffix="unused_brian_object",
            )

    def add_dependency(self, obj):
        """
        Add an object to the list of dependencies. Takes care of handling
        subgroups correctly (i.e., adds its parent object).

        Parameters
        ----------
        obj : `BrianObject`
            The object that this object depends on.
        """
        from brian2.groups.subgroup import Subgroup

        if isinstance(obj, Subgroup):
            self._dependencies.add(obj.source.id)
        else:
            self._dependencies.add(obj.id)

    def before_run(self, run_namespace):
        """
        Optional method to prepare the object before a run.

        Called by `Network.after_run` before the main simulation loop starts.
        """
        for codeobj in self._code_objects:
            codeobj.before_run()

    def after_run(self):
        """
        Optional method to do work after a run is finished.

        Called by `Network.after_run` after the main simulation loop terminated.
        """
        for codeobj in self._code_objects:
            codeobj.after_run()

    def run(self):
        for codeobj in self._code_objects:
            codeobj()

    contained_objects = property(
        fget=lambda self: self._contained_objects,
        doc="""
        The list of objects contained within the `BrianObject`.

        When a `BrianObject` is added to a `Network`, its contained objects will
        be added as well. This allows for compound objects which contain
        a mini-network structure.

        Note that this attribute cannot be set directly, you need to modify
        the underlying list, e.g. ``obj.contained_objects.extend([A, B])``.
        """,
    )

    code_objects = property(
        fget=lambda self: self._code_objects,
        doc="""
        The list of `CodeObject` contained within the `BrianObject`.

        TODO: more details.

        Note that this attribute cannot be set directly, you need to modify
        the underlying list, e.g. ``obj.code_objects.extend([A, B])``.
        """,
    )

    updaters = property(
        fget=lambda self: self._updaters,
        doc="""
        The list of `Updater` that define the runtime behaviour of this object.

        TODO: more details.

        Note that this attribute cannot be set directly, you need to modify
        the underlying list, e.g. ``obj.updaters.extend([A, B])``.
        """,
    )

    clock = property(
        fget=lambda self: self._clock,
        doc="""
        The `Clock` determining when the object should be updated.

        Note that this cannot be changed after the object is
        created.
        """,
    )

    def _set_active(self, val):
        val = bool(val)
        self._active = val
        for obj in self.contained_objects:
            obj.active = val

    active = property(
        fget=lambda self: self._active,
        fset=_set_active,
        doc="""
        Whether or not the object should be run.

        Inactive objects will not have their `update`
        method called in `Network.run`. Note that setting or
        unsetting the `active` attribute will set or unset
        it for all `contained_objects`.
        """,
    )

    def __repr__(self):
        classname = self.__class__.__name__
        description = (
            f"{classname}(clock={self._clock}, when={self.when}, "
            f"order={self.order}, name={self.name!r})"
        )
        return description

    # This is a repeat from Nameable.name, but we want to get the documentation
    # here again
    name = Nameable.name


def weakproxy_with_fallback(obj):
    """
    Attempts to create a `weakproxy` to the object, but falls back to the object if not possible.
    """
    try:
        return weakref.proxy(obj)
    except TypeError:
        return obj


def device_override(name):
    """
    Decorates a function/method to allow it to be overridden by the current `Device`.

    The ``name`` is the function name in the `Device` to use as an override if it exists.

    The returned function has an additional attribute ``original_function``
    which is a reference to the original, undecorated function.
    """

    def device_override_decorator(func):
        def device_override_decorated_function(*args, **kwds):
            from brian2.devices.device import get_device

            curdev = get_device()
            if hasattr(curdev, name):
                return getattr(curdev, name)(*args, **kwds)
            else:
                return func(*args, **kwds)

        device_override_decorated_function.original_function = func
        functools.update_wrapper(device_override_decorated_function, func)

        return device_override_decorated_function

    return device_override_decorator


class BrianObjectException(Exception):
    """
    High level exception that adds extra Brian-specific information to exceptions

    This exception should only be raised at a fairly high level in Brian code to
    pass information back to the user. It adds extra information about where a
    `BrianObject` was defined to better enable users to locate the source of
    problems.

    Parameters
    ----------

    message : str
        Additional error information to add to the original exception.
    brianobj : BrianObject
        The object that caused the error to happen.
    original_exception : Exception
        The original exception that was raised.
    """

    def __init__(self, message, brianobj):
        self._brian_message = message
        self._brian_objname = brianobj.name
        self._brian_objcreate = (
            "Object was created here (most recent call only, full details in "
            "debug log):\n" + brianobj._creation_stack
        )
        full_stack = "Object was created here:\n" + brianobj._full_creation_stack
        logger.diagnostic(
            "Error was encountered with object "
            f"'{self._brian_objname}':\n"
            f"{full_stack}"
        )

    def __str__(self):
        return (
            f"Error encountered with object named '{self._brian_objname}'.\n"
            f"{self._brian_objcreate}\n\n"
            f"{self._brian_message} "
            "(See above for original error message and traceback.)"
        )


def brian_object_exception(message, brianobj, original_exception):
    """
    Returns a `BrianObjectException` derived from the original exception.

    Creates a new class derived from the class of the original exception
    and `BrianObjectException`. This allows exception handling code to
    respond both to the original exception class and `BrianObjectException`.

    See `BrianObjectException` for arguments and notes.
    """

    raise NotImplementedError(
        "The brian_object_exception function is no longer used. "
        "Raise a BrianObjectException directly."
    )



"""
Module providing `NumpyCodeObject`.
"""

import sys
from collections.abc import Iterable

import numpy as np

from brian2.core.base import BrianObjectException
from brian2.core.functions import Function
from brian2.core.preferences import BrianPreference, prefs
from brian2.core.variables import (
    ArrayVariable,
    AuxiliaryVariable,
    DynamicArrayVariable,
    Subexpression,
)

from ...codeobject import CodeObject, check_compiler_kwds, constant_or_scalar
from ...generators.numpy_generator import NumpyCodeGenerator
from ...targets import codegen_targets
from ...templates import Templater

__all__ = ["NumpyCodeObject"]

# Preferences
prefs.register_preferences(
    "codegen.runtime.numpy",
    "Numpy runtime codegen preferences",
    discard_units=BrianPreference(
        default=False,
        docs="""
        Whether to change the namespace of user-specifed functions to remove
        units.
        """,
    ),
)


class LazyArange(Iterable):
    """
    A class that can be used as a `~numpy.arange` replacement (with an implied
    step size of 1) but does not actually create an array of values until
    necessary. It is somewhat similar to the ``range()`` function in Python 3,
    but does not use a generator. It is tailored to a special use case, the
    ``_vectorisation_idx`` variable in numpy templates, and not meant for
    general use. The ``_vectorisation_idx`` is used for stateless function
    calls such as ``rand()`` and for the numpy codegen target determines the
    number of values produced by such a call. This will often be the number of
    neurons or synapses, and this class avoids creating a new array of that size
    at every code object call when all that is needed is the *length* of the
    array.

    Examples
    --------
    >>> from brian2.codegen.runtime.numpy_rt.numpy_rt import LazyArange
    >>> ar = LazyArange(10)
    >>> len(ar)
    10
    >>> len(ar[:5])
    5
    >>> type(ar[:5])
    <class 'brian2.codegen.runtime.numpy_rt.numpy_rt.LazyArange'>
    >>> ar[5]
    5
    >>> for value in ar[3:7]:
    ...     print(value)
    ...
    3
    4
    5
    6
    >>> len(ar[np.array([1, 2, 3])])
    3
    """

    def __init__(self, stop, start=0, indices=None):
        self.start = start
        self.stop = stop
        self.indices = indices

    def __len__(self):
        if self.indices is None:
            return self.stop - self.start
        else:
            return len(self.indices)

    def __getitem__(self, item):
        if isinstance(item, slice):
            if self.indices is None:
                start, stop, step = item.start, item.stop, item.step
                if step not in [None, 1]:
                    raise NotImplementedError("Step should be 1")
                if start is None:
                    start = 0
                if stop is None:
                    stop = len(self)
                return LazyArange(
                    start=self.start + start, stop=min([self.start + stop, self.stop])
                )
            else:
                raise NotImplementedError("Cannot slice LazyArange with indices")
        elif isinstance(item, np.ndarray):
            if item.dtype == np.dtype(bool):
                item = np.nonzero(item)[0]  # convert boolean array into integers
            if len(item) == 0:
                return np.array([], dtype=np.int32)
            if np.min(item) < 0 or np.max(item) > len(self):
                raise IndexError("Indexing array contains out-of-bounds values")
            return LazyArange(start=self.start, stop=self.stop, indices=item)
        elif isinstance(item, int):
            if self.indices is None:
                index = self.start + item
                if index >= self.stop:
                    raise IndexError(index)
                return index
            else:
                return self.indices[item]
        else:
            raise TypeError("Can only index with integer, numpy array, or slice.")

    def __iter__(self):
        if self.indices is None:
            return iter(np.arange(self.start, self.stop))
        else:
            return iter(self.indices)

    # Allow conversion to a proper array with np.array(...)
    def __array__(self, dtype=None, copy=None):
        if copy is False:
            raise ValueError("LazyArange does not support copy=False")
        if self.indices is None:
            return np.arange(self.start, self.stop, dtype=dtype)
        else:
            return (self.indices + self.start).astype(dtype)

    # Allow basic arithmetics (used when shifting stuff for subgroups)
    def __add__(self, other):
        if isinstance(other, int):
            return LazyArange(start=self.start + other, stop=self.stop + other)
        else:
            return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, int):
            return LazyArange(start=self.start - other, stop=self.stop - other)
        else:
            return NotImplemented


class NumpyCodeObject(CodeObject):
    """
    Execute code using Numpy

    Default for Brian because it works on all platforms.
    """

    templater = Templater(
        "brian2.codegen.runtime.numpy_rt",
        ".py_",
        env_globals={"constant_or_scalar": constant_or_scalar},
    )
    generator_class = NumpyCodeGenerator
    class_name = "numpy"

    def __init__(
        self,
        owner,
        code,
        variables,
        variable_indices,
        template_name,
        template_source,
        compiler_kwds,
        name="numpy_code_object*",
    ):
        check_compiler_kwds(compiler_kwds, [], "numpy")
        from brian2.devices.device import get_device

        self.device = get_device()
        self.namespace = {
            "_owner": owner,
            # TODO: This should maybe go somewhere else
            "logical_not": np.logical_not,
        }
        CodeObject.__init__(
            self,
            owner,
            code,
            variables,
            variable_indices,
            template_name,
            template_source,
            compiler_kwds=compiler_kwds,
            name=name,
        )
        self.variables_to_namespace()

    @classmethod
    def is_available(cls):
        # no test necessary for numpy
        return True

    def variables_to_namespace(self):
        # Variables can refer to values that are either constant (e.g. dt)
        # or change every timestep (e.g. t). We add the values of the
        # constant variables here and add the names of non-constant variables
        # to a list

        # A list containing tuples of name and a function giving the value
        self.nonconstant_values = []

        for name, var in self.variables.items():
            if isinstance(var, (AuxiliaryVariable, Subexpression)):
                continue

            try:
                if not hasattr(var, "get_value"):
                    raise TypeError()
                value = var.get_value()
            except TypeError:
                # Either a dummy Variable without a value or a Function object
                if isinstance(var, Function):
                    impl = var.implementations[self.__class__].get_code(self.owner)
                    self.namespace[name] = impl
                else:
                    self.namespace[name] = var
                continue

            if isinstance(var, ArrayVariable):
                self.namespace[self.generator_class.get_array_name(var)] = value
                if var.scalar and var.constant:
                    self.namespace[name] = value[0]
            else:
                self.namespace[name] = value

            if isinstance(var, DynamicArrayVariable):
                dyn_array_name = self.generator_class.get_array_name(
                    var, access_data=False
                )
                self.namespace[dyn_array_name] = self.device.get_value(
                    var, access_data=False
                )

            # Also provide the Variable object itself in the namespace (can be
            # necessary for resize operations, for example)
            self.namespace[f"_var_{name}"] = var

            # There is one type of objects that we have to inject into the
            # namespace with their current value at each time step: dynamic
            # arrays that change in size during runs (i.e. not synapses but
            # e.g. the structures used in monitors)
            if isinstance(var, DynamicArrayVariable) and var.needs_reference_update:
                self.nonconstant_values.append(
                    (
                        self.generator_class.get_array_name(var, self.variables),
                        var.get_value,
                    )
                )

    def update_namespace(self):
        # update the values of the non-constant values in the namespace
        for name, func in self.nonconstant_values:
            self.namespace[name] = func()

    def compile_block(self, block):
        code = getattr(self.code, block, "").strip()
        if not code or "EMPTY_CODE_BLOCK" in code:
            return None
        return compile(code, "(string)", "exec")

    def run_block(self, block):
        compiled_code = self.compiled_code[block]
        if not compiled_code:
            return
        try:
            exec(compiled_code, self.namespace)
        except Exception as exc:
            code = getattr(self.code, block)
            message = (
                "An exception occured during the execution of the "
                f"'{block}' block of code object {self.name}.\n"
            )
            lines = code.split("\n")
            message += "The error was raised in the following line:\n"
            _, _, tb = sys.exc_info()
            tb = tb.tb_next  # Line in the code object's code
            message += f"{lines[tb.tb_lineno - 1]}\n"
            raise BrianObjectException(message, self.owner) from exc
        # output variables should land in the variable name _return_values
        if "_return_values" in self.namespace:
            return self.namespace["_return_values"]


codegen_targets.add(NumpyCodeObject)



"""
The unit system.
"""

# isort:skip_file

from .allunits import (
    # basic units
    pamp,
    namp,
    uamp,
    mamp,
    amp,
    kamp,
    Mamp,
    Gamp,
    Tamp,
    kelvin,
    kilogram,  # silly to have mkilogram, etc...
    pmetre,
    nmetre,
    umetre,
    mmetre,
    metre,
    kmetre,
    Mmetre,
    Gmetre,
    Tmetre,
    pmeter,
    nmeter,
    umeter,
    mmeter,
    meter,
    kmeter,
    Mmeter,
    Gmeter,
    Tmeter,
    cmetre,
    cmeter,  # quite commonly used
    psecond,
    nsecond,
    usecond,
    msecond,
    second,
    ksecond,
    Msecond,
    Gsecond,
    Tsecond,
    pmole,
    nmole,
    umole,
    mmole,
    mole,
    kmole,
    Mmole,
    Gmole,
    Tmole,
    # derived units
    pcoulomb,
    ncoulomb,
    ucoulomb,
    mcoulomb,
    coulomb,
    kcoulomb,
    Mcoulomb,
    Gcoulomb,
    Tcoulomb,
    pfarad,
    nfarad,
    ufarad,
    mfarad,
    farad,
    kfarad,
    Mfarad,
    Gfarad,
    Tfarad,
    pgram,
    ngram,
    ugram,
    mgram,
    gram,
    kgram,
    Mgram,
    Ggram,
    Tgram,
    pgramme,
    ngramme,
    ugramme,
    mgramme,
    gramme,
    kgramme,
    Mgramme,
    Ggramme,
    Tgramme,
    phertz,
    nhertz,
    uhertz,
    mhertz,
    hertz,
    khertz,
    Mhertz,
    Ghertz,
    Thertz,
    pjoule,
    njoule,
    ujoule,
    mjoule,
    joule,
    kjoule,
    Mjoule,
    Gjoule,
    Tjoule,
    pmolar,
    nmolar,
    umolar,
    mmolar,
    molar,
    kmolar,
    Mmolar,
    Gmolar,
    Tmolar,
    pliter,
    nliter,
    uliter,
    mliter,
    liter,
    kliter,
    Mliter,
    Gliter,
    Tliter,
    plitre,
    nlitre,
    ulitre,
    mlitre,
    litre,
    klitre,
    Mlitre,
    Glitre,
    Tlitre,
    ppascal,
    npascal,
    upascal,
    mpascal,
    pascal,
    kpascal,
    Mpascal,
    Gpascal,
    Tpascal,
    pohm,
    nohm,
    uohm,
    mohm,
    ohm,
    kohm,
    Mohm,
    Gohm,
    Tohm,
    psiemens,
    nsiemens,
    usiemens,
    msiemens,
    siemens,
    ksiemens,
    Msiemens,
    Gsiemens,
    Tsiemens,
    pvolt,
    nvolt,
    uvolt,
    mvolt,
    volt,
    kvolt,
    Mvolt,
    Gvolt,
    Tvolt,
    pwatt,
    nwatt,
    uwatt,
    mwatt,
    watt,
    kwatt,
    Mwatt,
    Gwatt,
    Twatt,
)
from .unitsafefunctions import *
from .unitsafefunctions import __all__ as unitsafefunctions_all

from .fundamentalunits import *
from .fundamentalunits import __all__ as fundamentalunits_all

from .stdunits import *
from .stdunits import __all__ as stdunits_all


__all__ = [
    "pamp",
    "namp",
    "uamp",
    "mamp",
    "amp",
    "kamp",
    "Mamp",
    "Gamp",
    "Tamp",
    "kelvin",
    "kilogram",
    "pmetre",
    "nmetre",
    "umetre",
    "mmetre",
    "metre",
    "kmetre",
    "Mmetre",
    "Gmetre",
    "Tmetre",
    "pmeter",
    "nmeter",
    "umeter",
    "mmeter",
    "meter",
    "kmeter",
    "Mmeter",
    "Gmeter",
    "Tmeter",
    "cmetre",
    "cmeter",
    "psecond",
    "nsecond",
    "usecond",
    "msecond",
    "second",
    "ksecond",
    "Msecond",
    "Gsecond",
    "Tsecond",
    "pmole",
    "nmole",
    "umole",
    "mmole",
    "mole",
    "kmole",
    "Mmole",
    "Gmole",
    "Tmole",
    # derived units
    "pcoulomb",
    "ncoulomb",
    "ucoulomb",
    "mcoulomb",
    "coulomb",
    "kcoulomb",
    "Mcoulomb",
    "Gcoulomb",
    "Tcoulomb",
    "pfarad",
    "nfarad",
    "ufarad",
    "mfarad",
    "farad",
    "kfarad",
    "Mfarad",
    "Gfarad",
    "Tfarad",
    "pgram",
    "ngram",
    "ugram",
    "mgram",
    "gram",
    "kgram",
    "Mgram",
    "Ggram",
    "Tgram",
    "pgramme",
    "ngramme",
    "ugramme",
    "mgramme",
    "gramme",
    "kgramme",
    "Mgramme",
    "Ggramme",
    "Tgramme",
    "phertz",
    "nhertz",
    "uhertz",
    "mhertz",
    "hertz",
    "khertz",
    "Mhertz",
    "Ghertz",
    "Thertz",
    "pjoule",
    "njoule",
    "ujoule",
    "mjoule",
    "joule",
    "kjoule",
    "Mjoule",
    "Gjoule",
    "Tjoule",
    "pmolar",
    "nmolar",
    "umolar",
    "mmolar",
    "molar",
    "kmolar",
    "Mmolar",
    "Gmolar",
    "Tmolar",
    "pliter",
    "nliter",
    "uliter",
    "mliter",
    "liter",
    "kliter",
    "Mliter",
    "Gliter",
    "Tliter",
    "plitre",
    "nlitre",
    "ulitre",
    "mlitre",
    "litre",
    "klitre",
    "Mlitre",
    "Glitre",
    "Tlitre",
    "ppascal",
    "npascal",
    "upascal",
    "mpascal",
    "pascal",
    "kpascal",
    "Mpascal",
    "Gpascal",
    "Tpascal",
    "pohm",
    "nohm",
    "uohm",
    "mohm",
    "ohm",
    "kohm",
    "Mohm",
    "Gohm",
    "Tohm",
    "psiemens",
    "nsiemens",
    "usiemens",
    "msiemens",
    "siemens",
    "ksiemens",
    "Msiemens",
    "Gsiemens",
    "Tsiemens",
    "pvolt",
    "nvolt",
    "uvolt",
    "mvolt",
    "volt",
    "kvolt",
    "Mvolt",
    "Gvolt",
    "Tvolt",
    "pwatt",
    "nwatt",
    "uwatt",
    "mwatt",
    "watt",
    "kwatt",
    "Mwatt",
    "Gwatt",
    "Twatt",
]
__all__.extend(unitsafefunctions_all)
__all__.extend(fundamentalunits_all)
__all__.extend(stdunits_all)



import itertools
import os
import pickle
import re
import shutil
import subprocess
import sys
import tempfile
from collections import defaultdict

import numpy

import brian2
from brian2.utils.stringtools import indent

__all__ = [
    "FeatureTest",
    "SpeedTest",
    "InaccuracyError",
    "Configuration",
    "run_feature_tests",
    "run_single_feature_test",
    "run_speed_tests",
    "DefaultConfiguration",
    "LocalConfiguration",
    "NumpyConfiguration",
    "CythonConfiguration",
    "CPPStandaloneConfiguration",
    "CPPStandaloneConfigurationOpenMP",
]


class InaccuracyError(AssertionError):
    def __init__(self, error, *args):
        self.error = error
        AssertionError.__init__(self, *args)


class BaseTest:
    """ """

    category = None  # a string with the category of features
    name = None  # a string with the particular feature name within the category
    tags = None  # a list of tags (strings) of features used
    # whether or not to allow the device to override the time: this can be used to remove the
    # compilation overheads on certain devices (but some tests might want to include this)
    allow_time_override = True

    @classmethod
    def fullname(cls):
        return f"{cls.category}: {cls.name}"

    def run(self):
        """
        Runs the feature test but do not return results (some devices may
        require an extra step before results are available).
        """
        raise NotImplementedError

    def timed_run(self, duration):
        """
        Do a timed run. This means that for RuntimeDevice it will run for defaultclock.dt before running for the
        rest of the duration. This means total run duration will be duration+defaultclock.dt.
        For standalone devices, this feature may or may not be implemented.
        """
        if isinstance(brian2.get_device(), brian2.devices.RuntimeDevice):
            brian2.run(brian2.defaultclock.dt, level=1)
            brian2.run(duration, level=1)
        else:
            brian2.run(duration, level=1)


class FeatureTest(BaseTest):
    """ """

    def results(self):
        """
        Return the results after a run call.
        """
        raise NotImplementedError

    def compare(self, maxrelerr, results_base, results_test):
        """
        Compare results from standard Brian run to another run.

        This method or `check` should be implemented.
        """
        raise NotImplementedError

    def check(self, maxrelerr, results):
        """
        Check results are valid (e.g. analytically).

        This method or `compare` should be implemented.
        """
        raise NotImplementedError

    def compare_arrays(self, maxrelerr, v_base, v_test):
        """
        Often you just want to compare the values of some arrays, this does that.
        """
        if isinstance(v_base, dict):
            for k in v_base:
                self.compare_arrays(maxrelerr, v_base[k], v_test[k])
        else:
            I = v_base != 0
            err = numpy.amax(numpy.abs(v_base[I] - v_test[I]) / v_base[I])
            if err > maxrelerr:
                raise InaccuracyError(err)
            if (v_test[-I] != 0).any():
                raise InaccuracyError(numpy.inf)


class SpeedTest(BaseTest):
    n_range = [1]
    n_label = "n"
    n_axis_log = True
    time_axis_log = True

    def __init__(self, n):
        self.n = n

    def results(self):
        return self.n

    def compare(self, maxrelerr, results_base, results_test):
        pass

    def check(self, maxrelerr, results):
        pass

    def __call__(self):
        return self


class Configuration:
    """ """

    name = None  # The name of this configuration

    def __init__(self, maximum_run_time=1e7 * brian2.second):
        maximum_run_time = float(maximum_run_time) * brian2.second
        self.maximum_run_time = maximum_run_time

    def before_run(self):
        pass

    def after_run(self):
        pass

    def get_last_run_time(self):
        """
        Implement this to overwrite the measured runtime (e.g. to remove overhead).
        """
        if hasattr(brian2.device, "_last_run_time"):
            return brian2.device._last_run_time
        raise NotImplementedError

    def get_last_run_completed_fraction(self):
        """
        Implement this to overwrite the amount of the last run that was completed (for devices that allow breaking
        early if the maximum run time is exceeded).
        """
        if hasattr(brian2.device, "_last_run_completed_fraction"):
            return brian2.device._last_run_completed_fraction
        return 1.0


class DefaultConfiguration(Configuration):
    name = "Default"

    def before_run(self):
        brian2.prefs.reset_to_defaults()
        brian2.set_device("runtime")


class LocalConfiguration(Configuration):
    name = "Local"

    def before_run(self):
        brian2.prefs.reset_to_defaults()
        brian2.set_device("runtime")
        brian2.prefs.load_preferences()


class NumpyConfiguration(Configuration):
    name = "Numpy"

    def before_run(self):
        brian2.prefs.reset_to_defaults()
        brian2.set_device("runtime")
        brian2.prefs.codegen.target = "numpy"


class CythonConfiguration(Configuration):
    name = "Cython"

    def before_run(self):
        brian2.prefs.reset_to_defaults()
        brian2.set_device("runtime")
        brian2.prefs.codegen.target = "cython"


class CPPStandaloneConfiguration(Configuration):
    name = "C++ standalone"

    def before_run(self):
        brian2.prefs.reset_to_defaults()
        brian2.set_device("cpp_standalone", build_on_run=False)

    def after_run(self):
        if os.path.exists("cpp_standalone"):
            shutil.rmtree("cpp_standalone")
        brian2.device.build(
            directory="cpp_standalone", compile=True, run=True, with_output=False
        )


class CPPStandaloneConfigurationOpenMP(Configuration):
    name = "C++ standalone (OpenMP)"

    def before_run(self):
        brian2.prefs.reset_to_defaults()
        brian2.set_device("cpp_standalone", build_on_run=False)
        brian2.prefs.devices.cpp_standalone.openmp_threads = 4

    def after_run(self):
        if os.path.exists("cpp_standalone"):
            shutil.rmtree("cpp_standalone")
        brian2.device.build(
            directory="cpp_standalone", compile=True, run=True, with_output=False
        )


def results(configuration, feature, n=None, maximum_run_time=1e7 * brian2.second):
    tempfilename = tempfile.mktemp("exception")
    if n is None:
        init_args = ""
    else:
        init_args = str(n)
    code_string = """
__file__ = '{fname}'
import brian2
from {config_module} import {config_name}
from {feature_module} import {feature_name}
configuration = {config_name}()
feature = {feature_name}({init_args})
import warnings, traceback, pickle, sys, os, time
warnings.simplefilter('ignore')
try:
    start_time = time.time()
    configuration.before_run()
    brian2.device._set_maximum_run_time({maximum_run_time})
    feature.run()
    configuration.after_run()
    results = feature.results()
    run_time = time.time()-start_time
    if feature.allow_time_override:
        try:
            run_time = configuration.get_last_run_time()
        except NotImplementedError:
            pass
    lrcf = configuration.get_last_run_completed_fraction()
    run_time = run_time/lrcf
    prof_info = brian2.magic_network.profiling_info
    new_prof_info = []
    for n, t in prof_info:
        new_prof_info.append((n, t/lrcf))
    f = open(r'{tempfname}', 'wb')
    pickle.dump((None, results, run_time, new_prof_info), f, -1)
    f.close()
except Exception, ex:
    #traceback.print_exc(file=sys.stdout)
    tb = traceback.format_exc()
    f = open(r'{tempfname}', 'wb')
    pickle.dump((tb, ex, 0.0, []), f, -1)
    f.close()
    """.format(
        config_module=configuration.__module__,
        config_name=configuration.__name__,
        feature_module=feature.__module__,
        feature_name=feature.__name__,
        tempfname=tempfilename,
        fname=__file__,
        init_args=init_args,
        maximum_run_time=float(maximum_run_time),
    )
    args = [sys.executable, "-c", code_string]
    # Run the example in a new process and make sure that stdout gets
    # redirected into the capture plugin
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    # sys.stdout.write(stdout)
    # sys.stderr.write(stderr)
    with open(tempfilename, "rb") as f:
        tb, res, runtime, profiling_info = pickle.load(f)
    return tb, res, runtime, profiling_info


def check_or_compare(feature, res, baseline, maxrelerr):
    feature = feature()
    try:
        feature.check(maxrelerr, res)
    except NotImplementedError:
        feature.compare(maxrelerr, baseline, res)


def run_single_feature_test(configuration, feature):
    return results(configuration, feature)


def run_feature_tests(
    configurations=None,
    feature_tests=None,
    strict=1e-5,
    tolerant=0.05,
    verbose=True,
    maximum_run_time=1e7 * brian2.second,
):
    if configurations is None:
        # some configurations to attempt to import
        try:
            import brian2genn.correctness_testing
        except:
            pass
        configurations = Configuration.__subclasses__()
    if feature_tests is None:
        feature_tests = FeatureTest.__subclasses__()
    if DefaultConfiguration in configurations:
        configurations.remove(DefaultConfiguration)
    configurations = [DefaultConfiguration] + configurations
    feature_tests.sort(key=lambda ft: ft.fullname())
    if verbose:
        print("Running feature tests")
        print("Configurations:", ", ".join(c.name for c in configurations))

    full_results = {}
    tag_results = defaultdict(lambda: defaultdict(list))
    for ft in feature_tests:
        baseline = None
        if verbose:
            print(f"{ft.fullname()}: [", end=" ")
        for configuration in configurations:
            txt = "OK"
            sym = "."
            exc = None
            tb, res, runtime, prof_info = results(
                configuration, ft, maximum_run_time=maximum_run_time
            )
            if isinstance(res, Exception):
                if isinstance(res, NotImplementedError):
                    sym = "N"
                    txt = "Not implemented"
                else:
                    sym = "E"
                    txt = "Error"
                if configuration is DefaultConfiguration:
                    raise res
            else:
                if configuration is DefaultConfiguration:
                    baseline = res
                try:
                    check_or_compare(ft, res, baseline, strict)
                except InaccuracyError as exc:
                    try:
                        check_or_compare(ft, res, baseline, tolerant)
                        sym = "I"
                        txt = f"Poor (error={100.0 * exc.error:.2f}%)"
                    except InaccuracyError as exc:
                        sym = "F"
                        txt = f"Fail (error={100.0 * exc.error:.2f}%)"
            sys.stdout.write(sym)
            full_results[configuration.name, ft.fullname()] = (
                sym,
                txt,
                exc,
                tb,
                runtime,
                prof_info,
            )
            for tag in ft.tags:
                tag_results[tag][configuration.name].append(
                    (sym, txt, exc, tb, runtime, prof_info)
                )
        if verbose:
            print("]")

    return FeatureTestResults(full_results, tag_results, configurations, feature_tests)


class FeatureTestResults:
    def __init__(self, full_results, tag_results, configurations, feature_tests):
        self.full_results = full_results
        self.tag_results = tag_results
        self.configurations = configurations
        self.feature_tests = feature_tests

    @property
    def test_table(self):
        table = []
        table.append(["Test"] + [c.name for c in self.configurations])
        curcat = ""

        for ft in self.feature_tests:
            cat = ft.category
            if cat != curcat:
                table.append([cat] + [""] * len(self.configurations))
                curcat = cat
            row = [ft.name]
            for configuration in self.configurations:
                sym, txt, exc, tb, runtime, prof_info = self.full_results[
                    configuration.name, ft.fullname()
                ]
                row.append(txt)
            table.append(row)
        return make_table(table)

    @property
    def tag_table(self):
        table = []
        table.append(["Tag"] + [c.name for c in self.configurations])
        tags = sorted(self.tag_results.keys())

        for tag in tags:
            row = [tag]
            for configuration in self.configurations:
                tag_res = self.tag_results[tag][configuration.name]
                syms = [sym for sym, txt, exc, tb, runtime, prof_info in tag_res]
                n = len(syms)
                okcount = sum(sym == "." for sym in syms)
                poorcount = sum(sym == "I" for sym in syms)
                failcount = sum(sym == "F" for sym in syms)
                errcount = sum(sym == "E" for sym in syms)
                nicount = sum(sym == "N" for sym in syms)
                if okcount == n:
                    txt = "OK"
                elif nicount == n:
                    txt = "Not implemented"
                elif errcount == n:
                    txt = "Unsupported"
                elif okcount + poorcount == n:
                    txt = f"Poor ({int(int(poorcount * 100.0 / n))}%)"
                elif okcount + poorcount + failcount == n:
                    txt = (
                        f"Fail: {int(failcount * 100.0 / n)}%"
                        f" (poor={int(poorcount * 100.0 / n)}%)"
                    )
                else:
                    txt = (
                        "Fail: OK={ok}%, Poor={poor}%, Fail={fail}%, NotImpl={ni}%"
                        " Error={err}%".format(
                            ok=int(okcount * 100.0 / n),
                            poor=int(poorcount * 100.0 / n),
                            fail=int(failcount * 100.0 / n),
                            err=int(errcount * 100.0 / n),
                            ni=int(nicount * 100.0 / n),
                        )
                    )
                row.append(txt)
            table.append(row)
        return make_table(table)

    @property
    def tables(self):
        r = ""
        s = "Feature test results"
        r += f"{s}\n{'-' * len(s)}\n\n{self.test_table}\n"
        s = "Tag results"
        r += f"{s}\n{'-' * len(s)}\n\n{self.tag_table}\n"
        return r

    @property
    def exceptions(self):
        exc_list = []
        for configuration in self.configurations:
            curconfig = []
            for ft in self.feature_tests:
                sym, txt, exc, tb, runtime, prof_info = self.full_results[
                    configuration.name, ft.fullname()
                ]
                if tb is not None:
                    curconfig.append((ft.fullname(), tb))
            if len(curconfig):
                exc_list.append((configuration.name, curconfig))
        if len(exc_list) == 0:
            return ""
        r = ""
        s = "Exceptions"
        r += f"{s}\n{'-' * len(s)}\n\n"
        for config_name, curconfig in exc_list:
            s = config_name
            r += f"{s}\n{'^' * len(s)}\n\n"
            for name, tb in curconfig:
                r += f"{name}::\n\n{indent(tb)}\n\n"
        return r

    @property
    def tables_and_exceptions(self):
        return f"{self.tables}\n{self.exceptions}"

    def __str__(self):
        return self.tables

    __repr__ = __str__


def run_speed_tests(
    configurations=None,
    speed_tests=None,
    run_twice=True,
    verbose=True,
    n_slice=slice(None),
    maximum_run_time=1e7 * brian2.second,
):
    if configurations is None:
        # some configurations to attempt to import
        try:
            import brian2genn.correctness_testing
        except:
            pass
        configurations = Configuration.__subclasses__()
    if speed_tests is None:
        speed_tests = SpeedTest.__subclasses__()
    speed_tests.sort(key=lambda ft: ft.fullname())
    if verbose:
        print("Running speed tests")
        print("Configurations:", ", ".join(c.name for c in configurations))

    full_results = {}
    tag_results = defaultdict(lambda: defaultdict(list))
    for ft in speed_tests:
        if verbose:
            print(f"{ft.fullname()}: ", end=" ")
        for n in ft.n_range[n_slice]:
            if verbose:
                print(f"n={int(n)} [", end=" ")
            for configuration in configurations:
                sym = "."
                for _ in range(1 + int(run_twice)):
                    tb, res, runtime, prof_info = results(
                        configuration, ft, n, maximum_run_time=maximum_run_time
                    )
                if isinstance(res, Exception):
                    if isinstance(res, NotImplementedError):
                        sym = "N"
                    else:
                        sym = "E"
                    if configuration is DefaultConfiguration:
                        raise res
                    runtime = numpy.NAN
                sys.stdout.write(sym)
                full_results[configuration.name, ft.fullname(), n, "All"] = runtime
                suffixtime = defaultdict(float)
                overheadstime = float(runtime)
                for codeobjname, proftime in prof_info:
                    # parts = codeobjname.split('_')
                    # parts = [part for part in parts if not re.match(r'\d+', part)]
                    # suffix = '_'.join(parts)
                    suffix = codeobjname
                    suffixtime[suffix] += proftime
                    overheadstime -= float(proftime)
                for suffix, proftime in list(suffixtime.items()):
                    full_results[configuration.name, ft.fullname(), n, suffix] = (
                        proftime
                    )
                full_results[configuration.name, ft.fullname(), n, "Overheads"] = (
                    overheadstime
                )
            if verbose:
                print("]", end=" ")
        if verbose:
            print()

    return SpeedTestResults(full_results, configurations, speed_tests)


class SpeedTestResults:
    def __init__(self, full_results, configurations, speed_tests):
        self.full_results = full_results
        self.configurations = configurations
        self.speed_tests = speed_tests

    def get_ns(self, fullname):
        L = [(cn, fn, n, s) for cn, fn, n, s in self.full_results if fn == fullname]
        confignames, fullnames, n, codeobjsuffixes = zip(*L)
        return numpy.array(sorted(list(set(n))))

    def get_codeobjsuffixes(self, fullname):
        L = [(cn, fn, n, s) for cn, fn, n, s in self.full_results if fn == fullname]
        confignames, fullnames, n, codeobjsuffixes = zip(*L)
        return set(codeobjsuffixes)

    def plot_all_tests(self, relative=False, profiling_minimum=1.0):
        if relative and profiling_minimum < 1:
            raise ValueError("Cannot use relative plots with profiling")
        import pylab

        for st in self.speed_tests:
            fullname = st.fullname()
            pylab.figure()
            ns = self.get_ns(fullname)
            codeobjsuffixes = self.get_codeobjsuffixes(fullname)
            codeobjsuffixes.remove("All")
            codeobjsuffixes.remove("Overheads")
            codeobjsuffixes = ["All", "Overheads"] + sorted(codeobjsuffixes)
            if relative or profiling_minimum == 1:
                codeobjsuffixes = ["All"]
            baseline = None
            havelabel = set()
            markerstyles_cycle = iter(
                itertools.cycle(["o", "s", "d", "v", "p", "h", "^", "<", ">"])
            )
            dashes = {}
            markerstyles = {}
            for isuffix, suffix in enumerate(codeobjsuffixes):
                cols = itertools.cycle(pylab.rcParams["axes.color_cycle"])
                for (iconfig, config), col in zip(enumerate(self.configurations), cols):
                    configname = config.name
                    runtimes = []
                    skip = True
                    for n in ns:
                        runtime = self.full_results.get(
                            (configname, fullname, n, "All"), numpy.nan
                        )
                        thistime = self.full_results.get(
                            (configname, fullname, n, suffix), numpy.nan
                        )
                        if float(thistime / runtime) >= profiling_minimum:
                            skip = False
                        runtimes.append(thistime)
                    if skip:
                        continue
                    runtimes = numpy.array(runtimes)
                    if relative:
                        if baseline is None:
                            baseline = runtimes
                        runtimes = baseline / runtimes
                    if suffix == "All":
                        lw = 2
                        label = configname
                    else:
                        lw = 1
                        label = suffix
                    plottable = sum(-numpy.isnan(runtimes[1:] + runtimes[:-1]))
                    if plottable:
                        if label in havelabel:
                            label = None
                        else:
                            havelabel.add(label)
                        dash = None
                        msty = None
                        if suffix != "All":
                            if suffix in dashes:
                                dash = dashes[suffix]
                                msty = markerstyles[suffix]
                            else:
                                j = len(dashes)
                                dash = (8, 2)
                                for b in bin(j)[2:]:
                                    if b == "0":
                                        dash = dash + (2, 2)
                                    else:
                                        dash = dash + (4, 2)
                                dashes[suffix] = dash
                                markerstyles[suffix] = msty = next(markerstyles_cycle)
                        line = pylab.plot(
                            ns,
                            runtimes,
                            lw=lw,
                            color=col,
                            marker=msty,
                            mec="none",
                            ms=8,
                            label=label,
                        )[0]
                        if dash is not None:
                            line.set_dashes(dash)
            pylab.title(fullname)
            pylab.legend(loc="best", fontsize="x-small", handlelength=8.0)
            pylab.xlabel(st.n_label)
            if st.n_axis_log:
                pylab.gca().set_xscale("log")
            if st.time_axis_log:
                pylab.gca().set_yscale("log")


# Code below auto generates restructured text tables, copied from:
# http://stackoverflow.com/questions/11347505/what-are-some-approaches-to-outputting-a-python-data-structure-to-restructuredte


def make_table(grid):
    max_cols = [
        max(out)
        for out in map(list, zip(*[[len(item) for item in row] for row in grid]))
    ]
    rst = table_div(max_cols, 1)

    for i, row in enumerate(grid):
        header_flag = False
        if i == 0 or i == len(grid) - 1:
            header_flag = True
        rst += normalize_row(row, max_cols)
        rst += table_div(max_cols, header_flag)
    return rst


def table_div(max_cols, header_flag=1):
    out = ""
    if header_flag == 1:
        style = "="
    else:
        style = "-"

    for max_col in max_cols:
        out += f"{max_col * style} "

    out += "\n"
    return out


def normalize_row(row, max_cols):
    r = ""
    for i, max_col in enumerate(max_cols):
        r += row[i] + (max_col - len(row[i]) + 1) * " "

    return f"{r}\n"



"""
Module implementing Brian's refractory mechanism.
"""

from brian2.units.allunits import second
from brian2.units.fundamentalunits import DIMENSIONLESS

from .equations import (
    BOOLEAN,
    DIFFERENTIAL_EQUATION,
    PARAMETER,
    Equations,
    Expression,
    SingleEquation,
)

__all__ = ["add_refractoriness"]


def check_identifier_refractory(identifier):
    """
    Check that the identifier is not using a name reserved for the refractory
    mechanism. The reserved names are `not_refractory`, `refractory`,
    `refractory_until`.

    Parameters
    ----------
    identifier : str
        The identifier to check.

    Raises
    ------
    ValueError
        If the identifier is a variable name used for the refractory mechanism.
    """
    if identifier in ("not_refractory", "refractory", "refractory_until"):
        raise SyntaxError(
            f"The name '{identifier}' is used in the refractory mechanism "
            " and should not be used as a variable "
            "name."
        )


Equations.register_identifier_check(check_identifier_refractory)


def add_refractoriness(eqs):
    """
    Extends a given set of equations with the refractory mechanism. New
    parameters are added and differential equations with the "unless refractory"
    flag are changed so that their right-hand side is 0 when the neuron is
    refractory (by multiplication with the ``not_refractory`` variable).

    Parameters
    ----------
    eqs : `Equations`
        The equations without refractory mechanism.

    Returns
    -------
    new_eqs : `Equations`
        New equations, with added parameters and changed differential
        equations having the "unless refractory" flag.
    """
    new_equations = []

    # replace differential equations having the active flag
    for eq in eqs.values():
        if eq.type == DIFFERENTIAL_EQUATION and "unless refractory" in eq.flags:
            # the only case where we have to change anything
            new_code = f"int(not_refractory)*({eq.expr.code})"
            new_equations.append(
                SingleEquation(
                    DIFFERENTIAL_EQUATION,
                    eq.varname,
                    eq.dim,
                    expr=Expression(new_code),
                    flags=eq.flags,
                )
            )
        else:
            new_equations.append(eq)

    # add new parameters
    new_equations.append(
        SingleEquation(PARAMETER, "not_refractory", DIMENSIONLESS, var_type=BOOLEAN)
    )
    new_equations.append(SingleEquation(PARAMETER, "lastspike", second.dim))

    return Equations(new_equations)



#!/usr/bin/env python
"""
Run the ``cuba.py`` example with OpenMP threads.
"""
from brian2 import *

set_device('cpp_standalone', directory='CUBA')
prefs.devices.cpp_standalone.openmp_threads = 4

taum = 20*ms
taue = 5*ms
taui = 10*ms
Vt = -50*mV
Vr = -60*mV
El = -49*mV

eqs = '''
dv/dt  = (ge+gi-(v-El))/taum : volt (unless refractory)
dge/dt = -ge/taue : volt (unless refractory)
dgi/dt = -gi/taui : volt (unless refractory)
'''

P = NeuronGroup(4000, eqs, threshold='v>Vt', reset='v = Vr', refractory=5*ms,
                method='exact')
P.v = 'Vr + rand() * (Vt - Vr)'
P.ge = 0*mV
P.gi = 0*mV

we = (60*0.27/10)*mV # excitatory synaptic weight (voltage)
wi = (-20*4.5/10)*mV # inhibitory synaptic weight
Ce = Synapses(P, P, on_pre='ge += we')
Ci = Synapses(P, P, on_pre='gi += wi')
Ce.connect('i<3200', p=0.02)
Ci.connect('i>=3200', p=0.02)

s_mon = SpikeMonitor(P)

run(1 * second)

plot(s_mon.t/ms, s_mon.i, ',k')
xlabel('Time (ms)')
ylabel('Neuron index')
show()



import ast
import inspect

from brian2.utils.stringtools import deindent, get_identifiers, indent

from .rendering import NodeRenderer

__all__ = [
    "AbstractCodeFunction",
    "abstract_code_from_function",
    "extract_abstract_code_functions",
    "substitute_abstract_code_functions",
]


class AbstractCodeFunction:
    """
    The information defining an abstract code function

    Has attributes corresponding to initialisation parameters

    Parameters
    ----------

    name : str
        The function name.
    args : list of str
        The arguments to the function.
    code : str
        The abstract code string consisting of the body of the function less
        the return statement.
    return_expr : str or None
        The expression returned, or None if there is nothing returned.
    """

    def __init__(self, name, args, code, return_expr):
        self.name = name
        self.args = args
        self.code = code
        self.return_expr = return_expr

    def __str__(self):
        s = (
            f"def {self.name}({', '.join(self.args)}):\n{indent(self.code)}\n    return"
            f" {self.return_expr}\n"
        )
        return s

    __repr__ = __str__


def abstract_code_from_function(func):
    """
    Converts the body of the function to abstract code

    Parameters
    ----------
    func : function, str or ast.FunctionDef
        The function object to convert. Note that the arguments to the
        function are ignored.

    Returns
    -------
    func : AbstractCodeFunction
        The corresponding abstract code function

    Raises
    ------
    SyntaxError
        If unsupported features are used such as if statements or indexing.
    """
    if callable(func):
        code = deindent(inspect.getsource(func))
        funcnode = ast.parse(code, mode="exec").body[0]
    elif isinstance(func, str):
        funcnode = ast.parse(func, mode="exec").body[0]
    elif func.__class__ is ast.FunctionDef:
        funcnode = func
    else:
        raise TypeError("Unsupported function type")

    if funcnode.args.vararg is not None:
        raise SyntaxError("No support for variable number of arguments")
    if funcnode.args.kwarg is not None:
        raise SyntaxError("No support for arbitrary keyword arguments")
    if len(funcnode.args.defaults):
        raise SyntaxError("No support for default values in functions")

    nodes = funcnode.body
    nr = NodeRenderer()
    lines = []
    return_expr = None
    for node in nodes:
        if node.__class__ is ast.Return:
            return_expr = nr.render_node(node.value)
            break
        else:
            lines.append(nr.render_node(node))
    abstract_code = "\n".join(lines)
    args = [arg.arg for arg in funcnode.args.args]
    name = funcnode.name
    return AbstractCodeFunction(name, args, abstract_code, return_expr)


def extract_abstract_code_functions(code):
    """
    Returns a set of abstract code functions from function definitions.

    Returns all functions defined at the top level and ignores any other
    code in the string.

    Parameters
    ----------
    code : str
        The code string defining some functions.

    Returns
    -------
    funcs : dict
        A mapping ``(name, func)`` for ``func`` an `AbstractCodeFunction`.
    """
    code = deindent(code)
    nodes = ast.parse(code, mode="exec").body
    funcs = {}
    for node in nodes:
        if node.__class__ is ast.FunctionDef:
            func = abstract_code_from_function(node)
            funcs[func.name] = func
    return funcs


class VarRewriter(ast.NodeTransformer):
    """
    Rewrites all variable names in names by prepending pre
    """

    def __init__(self, pre):
        self.pre = pre

    def visit_Name(self, node):
        return ast.Name(id=self.pre + node.id, ctx=node.ctx)

    def visit_Call(self, node):
        args = [self.visit(arg) for arg in node.args]
        return ast.Call(
            func=ast.Name(id=node.func.id, ctx=ast.Load()),
            args=args,
            keywords=[],
        )


class FunctionRewriter(ast.NodeTransformer):
    """
    Inlines a function call using temporary variables

    numcalls is the number of times the function rewriter has been called so
    far, this is used to make sure that when recursively inlining there is no
    name aliasing. The substitute_abstract_code_functions ensures that this is
    kept up to date between recursive runs.

    The pre attribute is the set of lines to be inserted above the currently
    being processed line, i.e. the inline code.

    The visit method returns the current line processed so that the function
    call is replaced with the output of the inlining.
    """

    def __init__(self, func, numcalls=0):
        self.func = func
        self.numcalls = numcalls
        self.pre = []
        self.suspend = False

    def visit_Call(self, node):
        # we suspend operations during an inlining operation, then resume
        # afterwards, see below, so we only ever try to expand one inline
        # function call at a time, i.e. no f(f(x)). This case is handled
        # by the recursion.
        if self.suspend:
            return node
        # We only work with the function we're provided
        if node.func.id != self.func.name:
            return node
        # Suspend while processing arguments (no recursion)
        self.suspend = True
        args = [self.visit(arg) for arg in node.args]
        self.suspend = False
        # The basename is used for function-local variables
        basename = f"_inline_{self.func.name}_{str(self.numcalls)}"
        # Assign all the function-local variables
        for argname, arg in zip(self.func.args, args):
            newpre = ast.Assign(
                targets=[ast.Name(id=f"{basename}_{argname}", ctx=ast.Store())],
                value=arg,
            )
            self.pre.append(newpre)
        # Rewrite the lines of code of the function using the names defined
        # above
        vr = VarRewriter(f"{basename}_")
        for funcline in ast.parse(self.func.code).body:
            self.pre.append(vr.visit(funcline))
        # And rewrite the return expression
        return_expr = vr.visit(ast.parse(self.func.return_expr, mode="eval").body)
        self.pre.append(
            ast.Assign(
                targets=[ast.Name(id=basename, ctx=ast.Store())], value=return_expr
            )
        )
        # Finally we replace the function call with the output of the inlining
        newnode = ast.Name(id=basename)
        self.numcalls += 1
        return newnode


def substitute_abstract_code_functions(code, funcs):
    """
    Performs inline substitution of all the functions in the code

    Parameters
    ----------
    code : str
        The abstract code to make inline substitutions into.
    funcs : list, dict or set of AbstractCodeFunction
        The function substitutions to use, note in the case of a dict, the
        keys are ignored and the function name is used.

    Returns
    -------
    code : str
        The code with inline substitutions performed.
    """
    if isinstance(funcs, (list, set)):
        newfuncs = dict()
        for f in funcs:
            newfuncs[f.name] = f
        funcs = newfuncs

    code = deindent(code)
    lines = ast.parse(code, mode="exec").body

    # This is a slightly nasty hack, but basically we just check by looking at
    # the existing identifiers how many inline operations have already been
    # performed by previous calls to this function
    ids = get_identifiers(code)
    funcstarts = {}
    for func in funcs.values():
        subids = {id for id in ids if id.startswith(f"_inline_{func.name}_")}
        subids = {id.replace(f"_inline_{func.name}_", "") for id in subids}
        alli = []
        for subid in subids:
            p = subid.find("_")
            if p > 0:
                subid = subid[:p]
            i = int(subid)
            alli.append(i)
        if len(alli) == 0:
            i = 0
        else:
            i = max(alli) + 1
        funcstarts[func.name] = i

    # Now we rewrite all the lines, replacing each line with a sequence of
    # lines performing the inlining
    newlines = []
    for line in lines:
        for func in funcs.values():
            rw = FunctionRewriter(func, funcstarts[func.name])
            line = rw.visit(line)
            newlines.extend(rw.pre)
            funcstarts[func.name] = rw.numcalls
        newlines.append(line)

    # Now we render to a code string
    nr = NodeRenderer()
    newcode = "\n".join(nr.render_node(line) for line in newlines)

    # We recurse until no changes in the code to ensure that all functions
    # are expanded if one function refers to another, etc.
    if newcode == code:
        return newcode
    else:
        return substitute_abstract_code_functions(newcode, funcs)
